(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('rxjs/operators'), require('rxjs'), require('@angular/core')) :
    typeof define === 'function' && define.amd ? define('@rx-angular/cdk', ['exports', 'rxjs/operators', 'rxjs', '@angular/core'], factory) :
    (global = typeof globalThis !== 'undefined' ? globalThis : global || self, factory((global['rx-angular'] = global['rx-angular'] || {}, global['rx-angular'].cdk = {}), global.rxjs.operators, global.rxjs, global.ng.core));
}(this, (function (exports, operators, rxjs, i0) { 'use strict';

    exports.RxNotificationKind = void 0;
    (function (RxNotificationKind) {
        RxNotificationKind["suspense"] = "suspense";
        RxNotificationKind["next"] = "next";
        RxNotificationKind["error"] = "error";
        RxNotificationKind["complete"] = "complete";
    })(exports.RxNotificationKind || (exports.RxNotificationKind = {}));

    /**
     * This Observable factory creates an Observable out of a static value or ObservableInput.
     *
     * @param o - the value to coerce
     */
    function coerceObservable(o) {
        return rxjs.isObservable(o) ? o : rxjs.of(o);
    }

    /**
     * This Observable factory creates an Observable out of a static value or ObservableInput.
     * It forwards only distinct values from distinct incoming Observables or values.
     * This comes in handy in any environment where you handle processing of incoming dynamic values and their state.
     *
     * Optionally you can pass a flatten strategy to get find grained control of the flattening process. E.g. mergeAll, switchAll
     *
     * @param o$ - The Observable to coerce and map to a Observable with distinct values
     * @param flattenOperator - determines the flattening strategy e.g. mergeAll, concatAll, exhaust, switchAll. default is switchAll
     */
    function coerceDistinctObservable(o$, flattenOperator) {
        flattenOperator = flattenOperator || operators.switchAll();
        return coerceObservable(o$).pipe(operators.distinctUntilChanged(), flattenOperator, operators.distinctUntilChanged());
    }

    /**
     * This operator maps an Observable out of a static value or ObservableInput.
     *
     */
    function coerceObservableWith() {
        return function (o$) { return operators.map(coerceObservable)(o$); };
    }

    /**
     * This operator takes an Observable of values ot Observables aof values and
     * It forwards only distinct values from distinct incoming Observables or values.
     * This comes in handy in any environment where you handle processing of incoming dynamic values and their state.
     *
     * Optionally you can pass a flatten strategy to get find grained control of the flattening process. E.g. mergeAll, switchAll
     *
     * @param flattenOperator - determines the flattening strategy e.g. mergeAll, concatAll, exhaust, switchAll. default is switchAll
     *
     */
    function coerceDistinctWith(flattenOperator) {
        flattenOperator = flattenOperator || operators.switchAll();
        return function (o$) { return o$.pipe(coerceObservableWith(), operators.distinctUntilChanged(), flattenOperator, operators.distinctUntilChanged()); };
    }

    /**
     * @internal
     *
     * A factory function returning an object to handle the process of merging Observable next notifications into one
     *   Observable. This API takes away the clumsy handling of static values and Observable, reduces the number of
     *   emissions by:
     * - only merging distinct Observables
     * - only emit distingt values of the merged result
     *
     * You can next a Observable of `U` multiple times and merge them into the Observable exposed under one optimized
     *   `values$`
     *
     */
    function hotFlatten(subjectFactory, flattenOperator) {
        var observablesSubject = subjectFactory ? subjectFactory() : new rxjs.Subject();
        flattenOperator = flattenOperator || operators.switchAll();
        var values$ = observablesSubject.pipe(coerceDistinctWith(flattenOperator));
        return {
            next: function (observable) {
                observablesSubject.next(observable);
            },
            values$: values$,
        };
    }

    function toRxErrorNotification(error, value) {
        return {
            kind: exports.RxNotificationKind.error,
            hasValue: !!value || false,
            value: value,
            complete: false,
            error: error || true,
        };
    }
    function toRxSuspenseNotification(value) {
        return {
            kind: exports.RxNotificationKind.suspense,
            hasValue: !!value || false,
            value: value,
            complete: false,
            error: false,
        };
    }
    function toRxCompleteNotification(value) {
        return {
            kind: exports.RxNotificationKind.complete,
            hasValue: !!value || false,
            value: value,
            complete: true,
            error: false,
        };
    }

    var notificationToRxNotification = function (notification) {
        var kind = rxJsToRxA[notification.kind];
        if (kind === exports.RxNotificationKind.error) {
            return toRxErrorNotification(notification.error);
        }
        if (kind === exports.RxNotificationKind.complete) {
            return toRxCompleteNotification(notification.value);
        }
        return Object.assign(Object.assign({}, notification), { error: false, complete: false, kind: kind });
    };
    var ɵ0 = notificationToRxNotification;
    var rxJsToRxA = {
        C: exports.RxNotificationKind.complete,
        E: exports.RxNotificationKind.error,
        N: exports.RxNotificationKind.next,
    };
    function rxMaterialize() {
        return function (o$) { return o$.pipe(operators.materialize(), 
        /*tap(({ kind, error }) => {
          if (kind === 'E') {
            console.error(error);
          }
        }),*/
        operators.map(notificationToRxNotification)); };
    }

    /**
     * @internal
     */
    function templateNotifier(withSuspense) {
        var observablesSubject = new rxjs.ReplaySubject(1);
        var firstRun = true;
        var latestNextValue;
        var wSuspense = function () { return withSuspense && withSuspense(); };
        var values$ = observablesSubject.pipe(operators.distinctUntilChanged(), 
        // handle null | undefined assignment and new Observable reset
        operators.map(function (observable$) {
            var isObs = rxjs.isObservable(observable$);
            // only pass through initial undefined value to filter out
            if (firstRun && (observable$ === undefined || observable$ === rxjs.NEVER)) {
                return wSuspense() ? rxjs.NEVER.pipe(operators.startWith(undefined)) : undefined;
            }
            var isNull = observable$ == null;
            if (isNull) {
                return rxjs.NEVER.pipe(operators.startWith(observable$));
            }
            return (isObs ? observable$ : rxjs.from(observable$)).pipe(function (o$) {
                if (!firstRun || (withSuspense && withSuspense())) {
                    return o$.pipe(operators.startWith(undefined));
                }
                return o$;
            });
        }), operators.tap(function () { return (firstRun = false); }), operators.filter(function (v) { return v !== undefined; }), operators.switchMap(function (o) {
            return o.pipe(operators.distinctUntilChanged(), rxMaterialize(), operators.map(function (notification) {
                latestNextValue =
                    notification.kind === exports.RxNotificationKind.next
                        ? notification.value
                        : latestNextValue;
                if (notification.kind === exports.RxNotificationKind.next &&
                    latestNextValue === undefined) {
                    return toRxSuspenseNotification(latestNextValue);
                }
                return Object.assign(Object.assign({}, notification), { value: latestNextValue });
            }));
        }));
        return {
            next: function (observable) {
                observablesSubject.next(observable);
            },
            values$: values$,
        };
    }

    /**
     * @internal
     *
     * @param value
     * @param strategy
     * @param workFactory
     * @param options
     */
    /*export function onStrategy<T>(
      value: T,
      strategy: RxStrategyCredentials,
      workFactory: (
        value: T,
        work: RxRenderWork,
        options: RxCoalescingOptions
      ) => void,
      options: RxCoalescingOptions = {}
    ): Observable<T> {
      return of(value).pipe(
        strategy.behavior(
          () => workFactory(value, strategy.work, options),
          options.scope || {}
        )
      );
    }*/
    /**
     * @internal
     *
     * @param value
     * @param strategy
     * @param workFactory
     * @param options
     * @param errorFactory
     */
    function onStrategy(value, strategy, workFactory, options, errorFactory) {
        if (options === void 0) { options = {}; }
        if (errorFactory === void 0) { errorFactory = function (e, v) { return [e, v]; }; }
        var error;
        return rxjs.of(value).pipe(strategy.behavior(function () {
            try {
                workFactory(value, strategy.work, options);
            }
            catch (e) {
                error = e;
            }
        }, options.scope || {}), operators.switchMap(function () { return error ? rxjs.throwError(errorFactory(error, value)) : rxjs.of(value); }));
    }

    /**
     * @internal
     *
     * A factory function returning an object to handle the process of turning strategy names into `RxStrategyCredentials`
     * You can next a strategy name as Observable or string and get an Observable of `RxStrategyCredentials`
     *
     * @param defaultStrategyName
     * @param strategies
     */
    function strategyHandling(defaultStrategyName, strategies) {
        var hotFlattened = hotFlatten(function () { return new rxjs.ReplaySubject(1); }, operators.switchAll());
        return {
            strategy$: hotFlattened.values$.pipe(operators.startWith(defaultStrategyName), nameToStrategyCredentials(strategies, defaultStrategyName), operators.share()),
            next: function (name) {
                hotFlattened.next(name);
            },
        };
    }
    /**
     * @internal
     */
    function nameToStrategyCredentials(strategies, defaultStrategyName) {
        return function (o$) { return o$.pipe(operators.map(function (name) { return name && Object.keys(strategies).includes(name)
            ? strategies[name]
            : strategies[defaultStrategyName]; })); };
    }

    /**
     * @internal
     *
     * A factory function returning an object to handle the process of switching templates by Notification channel.
     * You can next a Observable of `RxNotification` multiple times and merge them into the Observable exposed under `trigger$`
     *
     */
    function templateTriggerHandling() {
        var hotFlattened = hotFlatten(function () { return new rxjs.Subject(); }, operators.mergeAll());
        return {
            next: function (templateName) {
                hotFlattened.next(templateName);
            },
            trigger$: hotFlattened.values$.pipe(operators.share()),
        };
    }

    /*! *****************************************************************************
    Copyright (c) Microsoft Corporation.

    Permission to use, copy, modify, and/or distribute this software for any
    purpose with or without fee is hereby granted.

    THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
    REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
    AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
    INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
    LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
    OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
    PERFORMANCE OF THIS SOFTWARE.
    ***************************************************************************** */
    /* global Reflect, Promise */
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b)
                if (Object.prototype.hasOwnProperty.call(b, p))
                    d[p] = b[p]; };
        return extendStatics(d, b);
    };
    function __extends(d, b) {
        if (typeof b !== "function" && b !== null)
            throw new TypeError("Class extends value " + String(b) + " is not a constructor or null");
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    }
    var __assign = function () {
        __assign = Object.assign || function __assign(t) {
            for (var s, i = 1, n = arguments.length; i < n; i++) {
                s = arguments[i];
                for (var p in s)
                    if (Object.prototype.hasOwnProperty.call(s, p))
                        t[p] = s[p];
            }
            return t;
        };
        return __assign.apply(this, arguments);
    };
    function __rest(s, e) {
        var t = {};
        for (var p in s)
            if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
                t[p] = s[p];
        if (s != null && typeof Object.getOwnPropertySymbols === "function")
            for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
                if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                    t[p[i]] = s[p[i]];
            }
        return t;
    }
    function __decorate(decorators, target, key, desc) {
        var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
        if (typeof Reflect === "object" && typeof Reflect.decorate === "function")
            r = Reflect.decorate(decorators, target, key, desc);
        else
            for (var i = decorators.length - 1; i >= 0; i--)
                if (d = decorators[i])
                    r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
        return c > 3 && r && Object.defineProperty(target, key, r), r;
    }
    function __param(paramIndex, decorator) {
        return function (target, key) { decorator(target, key, paramIndex); };
    }
    function __metadata(metadataKey, metadataValue) {
        if (typeof Reflect === "object" && typeof Reflect.metadata === "function")
            return Reflect.metadata(metadataKey, metadataValue);
    }
    function __awaiter(thisArg, _arguments, P, generator) {
        function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
        return new (P || (P = Promise))(function (resolve, reject) {
            function fulfilled(value) { try {
                step(generator.next(value));
            }
            catch (e) {
                reject(e);
            } }
            function rejected(value) { try {
                step(generator["throw"](value));
            }
            catch (e) {
                reject(e);
            } }
            function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
            step((generator = generator.apply(thisArg, _arguments || [])).next());
        });
    }
    function __generator(thisArg, body) {
        var _ = { label: 0, sent: function () { if (t[0] & 1)
                throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
        return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function () { return this; }), g;
        function verb(n) { return function (v) { return step([n, v]); }; }
        function step(op) {
            if (f)
                throw new TypeError("Generator is already executing.");
            while (_)
                try {
                    if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done)
                        return t;
                    if (y = 0, t)
                        op = [op[0] & 2, t.value];
                    switch (op[0]) {
                        case 0:
                        case 1:
                            t = op;
                            break;
                        case 4:
                            _.label++;
                            return { value: op[1], done: false };
                        case 5:
                            _.label++;
                            y = op[1];
                            op = [0];
                            continue;
                        case 7:
                            op = _.ops.pop();
                            _.trys.pop();
                            continue;
                        default:
                            if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) {
                                _ = 0;
                                continue;
                            }
                            if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) {
                                _.label = op[1];
                                break;
                            }
                            if (op[0] === 6 && _.label < t[1]) {
                                _.label = t[1];
                                t = op;
                                break;
                            }
                            if (t && _.label < t[2]) {
                                _.label = t[2];
                                _.ops.push(op);
                                break;
                            }
                            if (t[2])
                                _.ops.pop();
                            _.trys.pop();
                            continue;
                    }
                    op = body.call(thisArg, _);
                }
                catch (e) {
                    op = [6, e];
                    y = 0;
                }
                finally {
                    f = t = 0;
                }
            if (op[0] & 5)
                throw op[1];
            return { value: op[0] ? op[1] : void 0, done: true };
        }
    }
    var __createBinding = Object.create ? (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        Object.defineProperty(o, k2, { enumerable: true, get: function () { return m[k]; } });
    }) : (function (o, m, k, k2) {
        if (k2 === undefined)
            k2 = k;
        o[k2] = m[k];
    });
    function __exportStar(m, o) {
        for (var p in m)
            if (p !== "default" && !Object.prototype.hasOwnProperty.call(o, p))
                __createBinding(o, m, p);
    }
    function __values(o) {
        var s = typeof Symbol === "function" && Symbol.iterator, m = s && o[s], i = 0;
        if (m)
            return m.call(o);
        if (o && typeof o.length === "number")
            return {
                next: function () {
                    if (o && i >= o.length)
                        o = void 0;
                    return { value: o && o[i++], done: !o };
                }
            };
        throw new TypeError(s ? "Object is not iterable." : "Symbol.iterator is not defined.");
    }
    function __read(o, n) {
        var m = typeof Symbol === "function" && o[Symbol.iterator];
        if (!m)
            return o;
        var i = m.call(o), r, ar = [], e;
        try {
            while ((n === void 0 || n-- > 0) && !(r = i.next()).done)
                ar.push(r.value);
        }
        catch (error) {
            e = { error: error };
        }
        finally {
            try {
                if (r && !r.done && (m = i["return"]))
                    m.call(i);
            }
            finally {
                if (e)
                    throw e.error;
            }
        }
        return ar;
    }
    /** @deprecated */
    function __spread() {
        for (var ar = [], i = 0; i < arguments.length; i++)
            ar = ar.concat(__read(arguments[i]));
        return ar;
    }
    /** @deprecated */
    function __spreadArrays() {
        for (var s = 0, i = 0, il = arguments.length; i < il; i++)
            s += arguments[i].length;
        for (var r = Array(s), k = 0, i = 0; i < il; i++)
            for (var a = arguments[i], j = 0, jl = a.length; j < jl; j++, k++)
                r[k] = a[j];
        return r;
    }
    function __spreadArray(to, from) {
        for (var i = 0, il = from.length, j = to.length; i < il; i++, j++)
            to[j] = from[i];
        return to;
    }
    function __await(v) {
        return this instanceof __await ? (this.v = v, this) : new __await(v);
    }
    function __asyncGenerator(thisArg, _arguments, generator) {
        if (!Symbol.asyncIterator)
            throw new TypeError("Symbol.asyncIterator is not defined.");
        var g = generator.apply(thisArg, _arguments || []), i, q = [];
        return i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i;
        function verb(n) { if (g[n])
            i[n] = function (v) { return new Promise(function (a, b) { q.push([n, v, a, b]) > 1 || resume(n, v); }); }; }
        function resume(n, v) { try {
            step(g[n](v));
        }
        catch (e) {
            settle(q[0][3], e);
        } }
        function step(r) { r.value instanceof __await ? Promise.resolve(r.value.v).then(fulfill, reject) : settle(q[0][2], r); }
        function fulfill(value) { resume("next", value); }
        function reject(value) { resume("throw", value); }
        function settle(f, v) { if (f(v), q.shift(), q.length)
            resume(q[0][0], q[0][1]); }
    }
    function __asyncDelegator(o) {
        var i, p;
        return i = {}, verb("next"), verb("throw", function (e) { throw e; }), verb("return"), i[Symbol.iterator] = function () { return this; }, i;
        function verb(n, f) { i[n] = o[n] ? function (v) { return (p = !p) ? { value: __await(o[n](v)), done: n === "return" } : f ? f(v) : v; } : f; }
    }
    function __asyncValues(o) {
        if (!Symbol.asyncIterator)
            throw new TypeError("Symbol.asyncIterator is not defined.");
        var m = o[Symbol.asyncIterator], i;
        return m ? m.call(o) : (o = typeof __values === "function" ? __values(o) : o[Symbol.iterator](), i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i);
        function verb(n) { i[n] = o[n] && function (v) { return new Promise(function (resolve, reject) { v = o[n](v), settle(resolve, reject, v.done, v.value); }); }; }
        function settle(resolve, reject, d, v) { Promise.resolve(v).then(function (v) { resolve({ value: v, done: d }); }, reject); }
    }
    function __makeTemplateObject(cooked, raw) {
        if (Object.defineProperty) {
            Object.defineProperty(cooked, "raw", { value: raw });
        }
        else {
            cooked.raw = raw;
        }
        return cooked;
    }
    ;
    var __setModuleDefault = Object.create ? (function (o, v) {
        Object.defineProperty(o, "default", { enumerable: true, value: v });
    }) : function (o, v) {
        o["default"] = v;
    };
    function __importStar(mod) {
        if (mod && mod.__esModule)
            return mod;
        var result = {};
        if (mod != null)
            for (var k in mod)
                if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k))
                    __createBinding(result, mod, k);
        __setModuleDefault(result, mod);
        return result;
    }
    function __importDefault(mod) {
        return (mod && mod.__esModule) ? mod : { default: mod };
    }
    function __classPrivateFieldGet(receiver, privateMap) {
        if (!privateMap.has(receiver)) {
            throw new TypeError("attempted to get private field on non-instance");
        }
        return privateMap.get(receiver);
    }
    function __classPrivateFieldSet(receiver, privateMap, value) {
        if (!privateMap.has(receiver)) {
            throw new TypeError("attempted to set private field on non-instance");
        }
        privateMap.set(receiver, value);
        return value;
    }

    var coalescingManager = createCoalesceManager();
    /*
     * createPropertiesWeakMap
     *
     * @param getDefaults: (o: O) => P
     * Example:
     *
     * export interface Properties {
     *   isCoalescing: boolean;
     * }
     *
     * const obj: object = {
     *   foo: 'bar',
     *   isCoalescing: 'weakMap version'
     * };
     *
     * const getDefaults = (ctx: object): Properties => ({isCoalescing: false});
     * const propsMap = createPropertiesWeakMap<object, Properties>(getDefaults);
     *
     * console.log('obj before:', obj);
     * // {foo: "bar", isCoalescing: "weakMap version"}
     * console.log('props before:', propsMap.getProps(obj));
     * // {isCoalescing: "weakMap version"}
     *
     * propsMap.setProps(obj, {isCoalescing: true});
     * console.log('obj after:', obj);
     * // {foo: "bar", isCoalescing: "weakMap version"}
     * console.log('props after:', propsMap.getProps(obj));
     * // {isCoalescing: "true"}
     * */
    function createPropertiesWeakMap(getDefaults) {
        var propertyMap = new WeakMap();
        return {
            getProps: getProperties,
            setProps: setProperties,
        };
        function getProperties(ctx) {
            var defaults = getDefaults(ctx);
            var propertiesPresent = propertyMap.get(ctx);
            var properties;
            if (propertiesPresent !== undefined) {
                properties = propertiesPresent;
            }
            else {
                properties = {};
                Object.entries(defaults).forEach(function (_a) {
                    var _b = __read(_a, 2), prop = _b[0], value = _b[1];
                    properties[prop] = hasKey(ctx, prop) ? ctx[prop] : value;
                });
                propertyMap.set(ctx, properties);
            }
            return properties;
        }
        function setProperties(ctx, props) {
            var properties = getProperties(ctx);
            Object.entries(props).forEach(function (_a) {
                var _b = __read(_a, 2), prop = _b[0], value = _b[1];
                properties[prop] = value;
            });
            propertyMap.set(ctx, properties);
            return properties;
        }
        function hasKey(ctx, property) {
            return ctx[property] != null;
        }
    }
    var ɵ0$1 = function (ctx) { return ({
        numCoalescingSubscribers: 0,
    }); };
    var coalescingContextPropertiesMap = createPropertiesWeakMap(ɵ0$1);
    /**
     * @describe createCoalesceManager
     *
     * returns a
     * Maintains a weak map of component references ans flags
     * them if the coalescing process is already started for them.
     *
     * Used in render aware internally.
     */
    function createCoalesceManager() {
        return {
            remove: removeWork,
            add: addWork,
            isCoalescing: isCoalescing,
        };
        // Increments the number of subscriptions in a scope e.g. a class instance
        function removeWork(scope) {
            var numCoalescingSubscribers = coalescingContextPropertiesMap.getProps(scope).numCoalescingSubscribers -
                1;
            coalescingContextPropertiesMap.setProps(scope, {
                numCoalescingSubscribers: numCoalescingSubscribers >= 0 ? numCoalescingSubscribers : 0,
            });
        }
        // Decrements the number of subscriptions in a scope e.g. a class instance
        function addWork(scope) {
            var numCoalescingSubscribers = coalescingContextPropertiesMap.getProps(scope).numCoalescingSubscribers +
                1;
            coalescingContextPropertiesMap.setProps(scope, {
                numCoalescingSubscribers: numCoalescingSubscribers,
            });
        }
        // Checks if anybody else is already coalescing atm
        function isCoalescing(scope) {
            return (coalescingContextPropertiesMap.getProps(scope).numCoalescingSubscribers >
                0);
        }
    }

    /**
     * @description
     * Limits the number of synchronous emitted a value from the source Observable to
     * one emitted value per
     *   [`AnimationFrame`](https://developer.mozilla.org/en-US/docs/Web/API/Window/requestAnimationFrame), then repeats
     *   this process for every tick of the browsers event loop.
     *
     * The coalesce operator is based on the [throttle](https://rxjs-dev.firebaseapp.com/api/operators/throttle) operator.
     * In addition to that is provides emitted values for the trailing end only, as well as maintaining a context to scope
     *   coalescing.
     *
     * @param {function(value: T): SubscribableOrPromise} durationSelector - A function
     * that receives a value from the source Observable, for computing the silencing
     * duration for each source value, returned as an Observable or a Promise.
     * It defaults to `requestAnimationFrame` as durationSelector.
     * @param {Object} config - A configuration object to define `leading` and `trailing` behavior and the context object.
     * Defaults to `{ leading: false, trailing: true }`. The default scoping is per subscriber.
     * @return {Observable<T>} An Observable that performs the coalesce operation to
     * limit the rate of emissions from the source.
     *
     * @usageNotes
     * Emit clicks at a rate of at most one click per second
     * ```typescript
     * import { fromEvent, animationFrames } from 'rxjs';
     * import { coalesce } from 'ngRx/component';
     *
     * const clicks = fromEvent(document, 'click');
     * const result = clicks.pipe(coalesce(ev => animationFrames));
     * result.subscribe(x => console.log(x));
     * ```
     */
    function coalesceWith(durationSelector, scope) {
        var _scope = scope || {};
        return function (source) {
            var o$ = new rxjs.Observable(function (observer) {
                var rootSubscription = new rxjs.Subscription();
                rootSubscription.add(source.subscribe(createInnerObserver(observer, rootSubscription)));
                return rootSubscription;
            });
            return o$;
            function createInnerObserver(outerObserver, rootSubscription) {
                var actionSubscription;
                var latestValue;
                var tryEmitLatestValue = function () {
                    if (actionSubscription) {
                        // We only decrement the number if it is greater than 0 (isCoalescing)
                        coalescingManager.remove(_scope);
                        if (!coalescingManager.isCoalescing(_scope)) {
                            outerObserver.next(latestValue);
                        }
                    }
                };
                return {
                    complete: function () {
                        tryEmitLatestValue();
                        outerObserver.complete();
                    },
                    error: function (error) { return outerObserver.error(error); },
                    next: function (value) {
                        latestValue = value;
                        if (!actionSubscription) {
                            // tslint:disable-next-line:no-unused-expression
                            coalescingManager.add(_scope);
                            actionSubscription = durationSelector.subscribe({
                                error: function (error) { return outerObserver.error(error); },
                                next: function () {
                                    tryEmitLatestValue();
                                    actionSubscription.unsubscribe();
                                    actionSubscription = undefined;
                                },
                                complete: function () {
                                    tryEmitLatestValue();
                                    actionSubscription = undefined;
                                },
                            });
                            rootSubscription.add(new rxjs.Subscription(function () {
                                tryEmitLatestValue();
                                if (actionSubscription) {
                                    actionSubscription.unsubscribe();
                                    actionSubscription = undefined;
                                }
                            }));
                        }
                    },
                };
            }
        };
    }

    function push(heap, node) {
        var index = heap.length;
        heap.push(node);
        siftUp(heap, node, index);
    }
    function peek(heap) {
        var first = heap[0];
        return first === undefined ? null : first;
    }
    function pop(heap) {
        var first = heap[0];
        if (first !== undefined) {
            var last = heap.pop();
            if (last !== first) {
                heap[0] = last;
                siftDown(heap, last, 0);
            }
            return first;
        }
        else {
            return null;
        }
    }
    function siftUp(heap, node, i) {
        var index = i;
        while (true) {
            // tslint:disable-next-line:no-bitwise
            var parentIndex = (index - 1) >>> 1;
            var parent = heap[parentIndex];
            if (parent !== undefined && compare(parent, node) > 0) {
                // The parent is larger. Swap positions.
                heap[parentIndex] = node;
                heap[index] = parent;
                index = parentIndex;
            }
            else {
                // The parent is smaller. Exit.
                return;
            }
        }
    }
    function siftDown(heap, node, i) {
        var index = i;
        var length = heap.length;
        while (index < length) {
            var leftIndex = (index + 1) * 2 - 1;
            var left = heap[leftIndex];
            var rightIndex = leftIndex + 1;
            var right = heap[rightIndex];
            // If the left or right node is smaller, swap with the smaller of those.
            if (left !== undefined && compare(left, node) < 0) {
                if (right !== undefined && compare(right, left) < 0) {
                    heap[index] = right;
                    heap[rightIndex] = node;
                    index = rightIndex;
                }
                else {
                    heap[index] = left;
                    heap[leftIndex] = node;
                    index = leftIndex;
                }
            }
            else if (right !== undefined && compare(right, node) < 0) {
                heap[index] = right;
                heap[rightIndex] = node;
                index = rightIndex;
            }
            else {
                // Neither child is smaller. Exit.
                return;
            }
        }
    }
    function compare(a, b) {
        // Compare sort index first, then task id.
        var diff = a.sortIndex - b.sortIndex;
        return diff !== 0 ? diff : a.id - b.id;
    }

    /**
     * Copyright (c) Facebook, Inc. and its affiliates.
     *
     * This source code is licensed under the MIT license found in the
     * LICENSE file in the root directory of this source tree.
     *
     * @flow
     */
    // TODO: Use symbols?
    var NoPriority = 0;
    var ImmediatePriority = 1;
    var UserBlockingPriority = 2;
    var NormalPriority = 3;
    var LowPriority = 4;
    var IdlePriority = 5;

    var enableIsInputPending = false;

    // see https://raw.githubusercontent.com/facebook/react/master/packages/scheduler/src/forks/SchedulerDOM.js
    var getCurrentTime;
    var hasPerformanceNow = typeof performance === 'object' && typeof performance.now === 'function';
    if (hasPerformanceNow) {
        var localPerformance_1 = performance;
        getCurrentTime = function () { return localPerformance_1.now(); };
    }
    else {
        var localDate_1 = Date;
        var initialTime_1 = localDate_1.now();
        getCurrentTime = function () { return localDate_1.now() - initialTime_1; };
    }
    // Max 31 bit integer. The max integer size in V8 for 32-bit systems.
    // Math.pow(2, 30) - 1
    // 0b111111111111111111111111111111
    var maxSigned31BitInt = 1073741823;
    // Times out immediately
    var IMMEDIATE_PRIORITY_TIMEOUT = -1;
    // Eventually times out
    var USER_BLOCKING_PRIORITY_TIMEOUT = 250;
    var NORMAL_PRIORITY_TIMEOUT = 5000;
    var LOW_PRIORITY_TIMEOUT = 10000;
    // Never times out
    var IDLE_PRIORITY_TIMEOUT = maxSigned31BitInt;
    // Tasks are stored on a min heap
    var taskQueue = [];
    var timerQueue = [];
    // Incrementing id counter. Used to maintain insertion order.
    var taskIdCounter = 1;
    // Pausing the scheduler is useful for debugging.
    var isSchedulerPaused = false;
    var currentTask = null;
    var currentPriorityLevel = NormalPriority;
    // This is set while performing work, to prevent re-entrancy.
    var isPerformingWork = false;
    var isHostCallbackScheduled = false;
    var isHostTimeoutScheduled = false;
    // Capture local references to native APIs, in case a polyfill overrides them.
    var setTimeout = i0.ɵglobal.setTimeout;
    var clearTimeout = i0.ɵglobal.clearTimeout;
    if (typeof console !== 'undefined') {
        // TODO: Scheduler no longer requires these methods to be polyfilled. But
        // maybe we want to continue warning if they don't exist, to preserve the
        // option to rely on it in the future?
        var requestAnimationFrame = i0.ɵglobal.requestAnimationFrame;
        var cancelAnimationFrame = i0.ɵglobal.cancelAnimationFrame;
        if (typeof requestAnimationFrame !== 'function') {
            // Using console['error'] to evade Babel and ESLint
            console['error']("This browser doesn't support requestAnimationFrame. " +
                'Make sure that you load a ' +
                'polyfill in older browsers. https://reactjs.org/link/react-polyfills');
        }
        if (typeof cancelAnimationFrame !== 'function') {
            // Using console['error'] to evade Babel and ESLint
            console['error']("This browser doesn't support cancelAnimationFrame. " +
                'Make sure that you load a ' +
                'polyfill in older browsers. https://reactjs.org/link/react-polyfills');
        }
    }
    function advanceTimers(currentTime) {
        // Check for tasks that are no longer delayed and add them to the queue.
        var timer = peek(timerQueue);
        while (timer !== null) {
            if (timer.callback === null) {
                // Timer was cancelled.
                pop(timerQueue);
            }
            else if (timer.startTime <= currentTime) {
                // Timer fired. Transfer to the task queue.
                pop(timerQueue);
                timer.sortIndex = timer.expirationTime;
                push(taskQueue, timer);
            }
            else {
                // Remaining timers are pending.
                return;
            }
            timer = peek(timerQueue);
        }
    }
    function handleTimeout(currentTime) {
        isHostTimeoutScheduled = false;
        advanceTimers(currentTime);
        if (!isHostCallbackScheduled) {
            if (peek(taskQueue) !== null) {
                isHostCallbackScheduled = true;
                requestHostCallback(flushWork);
            }
            else {
                var firstTimer = peek(timerQueue);
                if (firstTimer !== null) {
                    requestHostTimeout(handleTimeout, firstTimer.startTime - currentTime);
                }
            }
        }
    }
    function flushWork(hasTimeRemaining, initialTime) {
        // We'll need a host callback the next time work is scheduled.
        isHostCallbackScheduled = false;
        if (isHostTimeoutScheduled) {
            // We scheduled a timeout but it's no longer needed. Cancel it.
            isHostTimeoutScheduled = false;
            cancelHostTimeout();
        }
        isPerformingWork = true;
        var previousPriorityLevel = currentPriorityLevel;
        try {
            return workLoop(hasTimeRemaining, initialTime);
        }
        finally {
            currentTask = null;
            currentPriorityLevel = previousPriorityLevel;
            isPerformingWork = false;
        }
    }
    function workLoop(hasTimeRemaining, initialTime) {
        var currentTime = initialTime;
        advanceTimers(currentTime);
        currentTask = peek(taskQueue);
        while (currentTask !== null && !isSchedulerPaused) {
            if (currentTask.expirationTime > currentTime &&
                (!hasTimeRemaining || shouldYieldToHost())) {
                // This currentTask hasn't expired, and we've reached the deadline.
                break;
            }
            var callback = currentTask.callback;
            if (typeof callback === 'function') {
                currentTask.callback = null;
                currentPriorityLevel = currentTask.priorityLevel;
                var didUserCallbackTimeout = currentTask.expirationTime <= currentTime;
                var continuationCallback = callback(didUserCallbackTimeout);
                currentTime = getCurrentTime();
                if (typeof continuationCallback === 'function') {
                    currentTask.callback = continuationCallback;
                }
                else {
                    if (currentTask === peek(taskQueue)) {
                        pop(taskQueue);
                    }
                }
                advanceTimers(currentTime);
            }
            else {
                pop(taskQueue);
            }
            currentTask = peek(taskQueue);
        }
        // Return whether there's additional work
        if (currentTask !== null) {
            return true;
        }
        else {
            var firstTimer = peek(timerQueue);
            if (firstTimer !== null) {
                requestHostTimeout(handleTimeout, firstTimer.startTime - currentTime);
            }
            return false;
        }
    }
    function runWithPriority(priorityLevel, eventHandler) {
        switch (priorityLevel) {
            case ImmediatePriority:
            case UserBlockingPriority:
            case NormalPriority:
            case LowPriority:
            case IdlePriority:
                break;
            default:
                priorityLevel = NormalPriority;
        }
        var previousPriorityLevel = currentPriorityLevel;
        currentPriorityLevel = priorityLevel;
        try {
            return eventHandler();
        }
        finally {
            currentPriorityLevel = previousPriorityLevel;
        }
    }
    function next(eventHandler) {
        var priorityLevel;
        switch (currentPriorityLevel) {
            case ImmediatePriority:
            case UserBlockingPriority:
            case NormalPriority:
                // Shift down to normal priority
                priorityLevel = NormalPriority;
                break;
            default:
                // Anything lower than normal priority should remain at the current level.
                priorityLevel = currentPriorityLevel;
                break;
        }
        var previousPriorityLevel = currentPriorityLevel;
        currentPriorityLevel = priorityLevel;
        try {
            return eventHandler();
        }
        finally {
            currentPriorityLevel = previousPriorityLevel;
        }
    }
    function wrapCallback(callback) {
        var _this = this;
        var parentPriorityLevel = currentPriorityLevel;
        return function () {
            // This is a fork of runWithPriority, inlined for performance.
            var previousPriorityLevel = currentPriorityLevel;
            currentPriorityLevel = parentPriorityLevel;
            try {
                // @ts-ignore
                return callback.apply(_this, arguments);
            }
            finally {
                currentPriorityLevel = previousPriorityLevel;
            }
        };
    }
    function scheduleCallback(priorityLevel, callback, options) {
        var currentTime = getCurrentTime();
        var startTime;
        if (typeof options === 'object' && options !== null) {
            var delay = options.delay;
            if (typeof delay === 'number' && delay > 0) {
                startTime = currentTime + delay;
            }
            else {
                startTime = currentTime;
            }
        }
        else {
            startTime = currentTime;
        }
        var timeout;
        switch (priorityLevel) {
            case ImmediatePriority:
                timeout = IMMEDIATE_PRIORITY_TIMEOUT;
                break;
            case UserBlockingPriority:
                timeout = USER_BLOCKING_PRIORITY_TIMEOUT;
                break;
            case IdlePriority:
                timeout = IDLE_PRIORITY_TIMEOUT;
                break;
            case LowPriority:
                timeout = LOW_PRIORITY_TIMEOUT;
                break;
            case NormalPriority:
            default:
                timeout = NORMAL_PRIORITY_TIMEOUT;
                break;
        }
        var expirationTime = startTime + timeout;
        var newTask = {
            id: taskIdCounter++,
            callback: callback,
            priorityLevel: priorityLevel,
            startTime: startTime,
            expirationTime: expirationTime,
            sortIndex: -1,
        };
        if (startTime > currentTime) {
            // This is a delayed task.
            newTask.sortIndex = startTime;
            push(timerQueue, newTask);
            if (peek(taskQueue) === null && newTask === peek(timerQueue)) {
                // All tasks are delayed, and this is the task with the earliest delay.
                if (isHostTimeoutScheduled) {
                    // Cancel an existing timeout.
                    cancelHostTimeout();
                }
                else {
                    isHostTimeoutScheduled = true;
                }
                // Schedule a timeout.
                requestHostTimeout(handleTimeout, startTime - currentTime);
            }
        }
        else {
            newTask.sortIndex = expirationTime;
            push(taskQueue, newTask);
            // Schedule a host callback, if needed. If we're already performing work,
            // wait until the next time we yield.
            if (!isHostCallbackScheduled && !isPerformingWork) {
                isHostCallbackScheduled = true;
                requestHostCallback(flushWork);
            }
        }
        return newTask;
    }
    function pauseExecution() {
        isSchedulerPaused = true;
    }
    function continueExecution() {
        isSchedulerPaused = false;
        if (!isHostCallbackScheduled && !isPerformingWork) {
            isHostCallbackScheduled = true;
            requestHostCallback(flushWork);
        }
    }
    function getFirstCallbackNode() {
        return peek(taskQueue);
    }
    function cancelCallback(task) {
        // Null out the callback to indicate the task has been canceled. (Can't
        // remove from the queue because you can't remove arbitrary nodes from an
        // array based heap, only the first one.)
        task.callback = null;
    }
    function getCurrentPriorityLevel() {
        return currentPriorityLevel;
    }
    var isMessageLoopRunning = false;
    var scheduledHostCallback = null;
    var taskTimeoutID = -1;
    // Scheduler periodically yields in case there is other work on the main
    // thread, like user events. By default, it yields multiple times per frame.
    // It does not attempt to align with frame boundaries, since most tasks don't
    // need to be frame aligned; for those that do, use requestAnimationFrame.
    var yieldInterval = 16;
    var deadline = 0;
    // TODO: Make this configurable
    // TODO: Adjust this based on priority?
    var maxYieldInterval = 300;
    var needsPaint = false;
    function shouldYieldToHost() {
        if (enableIsInputPending &&
            navigator !== undefined &&
            navigator.scheduling !== undefined &&
            navigator.scheduling.isInputPending !== undefined) {
            var scheduling = navigator.scheduling;
            var currentTime = getCurrentTime();
            if (currentTime >= deadline) {
                // There's no time left. We may want to yield control of the main
                // thread, so the browser can perform high priority tasks. The main ones
                // are painting and user input. If there's a pending paint or a pending
                // input, then we should yield. But if there's neither, then we can
                // yield less often while remaining responsive. We'll eventually yield
                // regardless, since there could be a pending paint that wasn't
                // accompanied by a call to `requestPaint`, or other main thread tasks
                // like network events.
                if (needsPaint || scheduling.isInputPending()) {
                    // There is either a pending paint or a pending input.
                    return true;
                }
                // There's no pending input. Only yield if we've reached the max
                // yield interval.
                return currentTime >= maxYieldInterval;
            }
            else {
                // There's still time left in the frame.
                return false;
            }
        }
        else {
            // `isInputPending` is not available. Since we have no way of knowing if
            // there's pending input, always yield at the end of the frame.
            return getCurrentTime() >= deadline;
        }
    }
    function requestPaint() {
        if (enableIsInputPending &&
            navigator !== undefined &&
            navigator.scheduling !== undefined &&
            navigator.scheduling.isInputPending !== undefined) {
            needsPaint = true;
        }
        // Since we yield every frame regardless, `requestPaint` has no effect.
    }
    function forceFrameRate(fps) {
        if (fps < 0 || fps > 125) {
            // Using console['error'] to evade Babel and ESLint
            console['error']('forceFrameRate takes a positive int between 0 and 125, ' +
                'forcing frame rates higher than 125 fps is not supported');
            return;
        }
        if (fps > 0) {
            yieldInterval = Math.floor(1000 / fps);
        }
        else {
            // reset the framerate
            yieldInterval = 5;
        }
    }
    var performWorkUntilDeadline = function () {
        if (scheduledHostCallback !== null) {
            var currentTime = getCurrentTime();
            // Yield after `yieldInterval` ms, regardless of where we are in the vsync
            // cycle. This means there's always time remaining at the beginning of
            // the message event.
            deadline = currentTime + yieldInterval;
            var hasTimeRemaining = true;
            // If a scheduler task throws, exit the current browser task so the
            // error can be observed.
            //
            // Intentionally not using a try-catch, since that makes some debugging
            // techniques harder. Instead, if `scheduledHostCallback` errors, then
            // `hasMoreWork` will remain true, and we'll continue the work loop.
            var hasMoreWork = true;
            try {
                hasMoreWork = scheduledHostCallback(hasTimeRemaining, currentTime);
            }
            finally {
                if (hasMoreWork) {
                    // If there's more work, schedule the next message event at the end
                    // of the preceding one.
                    // tslint:disable-next-line: no-unused-expression
                    port && port.postMessage(null);
                }
                else {
                    isMessageLoopRunning = false;
                    scheduledHostCallback = null;
                }
            }
        }
        else {
            isMessageLoopRunning = false;
        }
        // Yielding to the browser will give it a chance to paint, so we can
        // reset this.
        needsPaint = false;
    };
    var ɵ0$2 = performWorkUntilDeadline;
    var channel = i0.ɵglobal.MessageChannel && new i0.ɵglobal.MessageChannel();
    var port = channel && channel.port2;
    if (channel) {
        channel.port1.onmessage = performWorkUntilDeadline;
    }
    function requestHostCallback(callback) {
        scheduledHostCallback = callback;
        if (!isMessageLoopRunning) {
            isMessageLoopRunning = true;
            // tslint:disable-next-line: no-unused-expression
            port && port.postMessage(null);
        }
    }
    function requestHostTimeout(callback, ms) {
        taskTimeoutID = setTimeout(function () {
            callback(getCurrentTime());
        }, ms);
    }
    function cancelHostTimeout() {
        clearTimeout(taskTimeoutID);
        taskTimeoutID = -1;
    }
    var _requestPaint = requestPaint;

    forceFrameRate(60);
    var ɵ0$3 = function (cdRef) { return cdRef.detectChanges(); }, ɵ1 = function (work, scope) {
        return function (o$) { return o$.pipe(scheduleOnQueue(work, { priority: NoPriority, scope: scope })); };
    };
    var noPriorityStrategy = {
        name: 'noPriority',
        work: ɵ0$3,
        behavior: ɵ1,
    };
    var ɵ2 = function (cdRef) { return cdRef.detectChanges(); }, ɵ3 = function (work, scope) {
        return function (o$) { return o$.pipe(scheduleOnQueue(work, {
            priority: ImmediatePriority,
            scope: scope,
        })); };
    };
    var immediateStrategy = {
        name: 'immediate',
        work: ɵ2,
        behavior: ɵ3,
    };
    var ɵ4 = function (cdRef) { return cdRef.detectChanges(); }, ɵ5 = function (work, scope) {
        return function (o$) { return o$.pipe(scheduleOnQueue(work, {
            priority: UserBlockingPriority,
            scope: scope,
        })); };
    };
    var userBlockingStrategy = {
        name: 'userBlocking',
        work: ɵ4,
        behavior: ɵ5,
    };
    var ɵ6 = function (cdRef) { return cdRef.detectChanges(); }, ɵ7 = function (work, scope) {
        return function (o$) { return o$.pipe(scheduleOnQueue(work, { priority: NormalPriority, scope: scope })); };
    };
    var normalStrategy = {
        name: 'normal',
        work: ɵ6,
        behavior: ɵ7,
    };
    var ɵ8 = function (cdRef) { return cdRef.detectChanges(); }, ɵ9 = function (work, scope) {
        return function (o$) { return o$.pipe(scheduleOnQueue(work, { priority: LowPriority, scope: scope })); };
    };
    var lowStrategy = {
        name: 'low',
        work: ɵ8,
        behavior: ɵ9,
    };
    var ɵ10 = function (cdRef) { return cdRef.detectChanges(); }, ɵ11 = function (work, scope) {
        return function (o$) { return o$.pipe(scheduleOnQueue(work, { priority: IdlePriority, scope: scope })); };
    };
    var idleStrategy = {
        name: 'idle',
        work: ɵ10,
        behavior: ɵ11,
    };
    function scheduleOnQueue(work, options) {
        return function (o$) { return o$.pipe(operators.filter(function () { return !coalescingManager.isCoalescing(options.scope); }), operators.switchMap(function (v) { return new rxjs.Observable(function (subscriber) {
            coalescingManager.add(options.scope);
            var task = scheduleCallback(options.priority, function () {
                work();
                coalescingManager.remove(options.scope);
                subscriber.next(v);
            }, { delay: options.delay });
            return function () {
                coalescingManager.remove(options.scope);
                cancelCallback(task);
            };
        }).pipe(operators.mapTo(v)); })); };
    }
    var RX_CONCURRENT_STRATEGIES = {
        noPriority: noPriorityStrategy,
        immediate: immediateStrategy,
        userBlocking: userBlockingStrategy,
        normal: normalStrategy,
        low: lowStrategy,
        idle: idleStrategy,
    };

    function getZoneUnPatchedApi(name, target) {
        if (target === void 0) { target = i0.ɵglobal; }
        var _a;
        return (_a = target['__zone_symbol__' + name]) !== null && _a !== void 0 ? _a : target[name];
    }

    /**
     * This file provides unpatched versions of APIs patched in the following file: https://github.com/angular/angular/blob/master/packages/zone.js/lib/browser/browser.ts
     */
    /**
     * This function is a zone un-patched implementation of Window#queueMicrotask() method.
     * It is which is exposed on the Window or Worker interface,
     * queues a microtask to be executed at a safe time prior to control returning to the browser's event loop.
     * The microtask is a short function which will run after the current task has completed its
     * work and when there is no other code waiting to be run before control of the execution context is returned to the browser's event loop.
     */
    function queueMicrotask() {
        return getZoneUnPatchedApi('queueMicrotask');
    }
    var Promise$1 = getZoneUnPatchedApi('Promise');
    /**
     * requestAnimationFrame
     *
     * @description
     *
     * This function is a zone un-patched implementation of Window#requestAnimationFrame() method
     *
     * The requestAnimationFrame() method calls a function or evaluates an expression on the next animationFrame.
     * The requestAnimationFrame() method will not continue calling the function after executed once.
     * The ID value returned by requestAnimationFrame() is used as the parameter for the cancelAnimationFrame() method.
     *
     * requestAnimationFrame(cb, ms);
     *
     * @param cb - Required. The function that will be executed
     *
     */
    function requestAnimationFrame$1(cb) {
        return getZoneUnPatchedApi('requestAnimationFrame')(cb);
    }
    /**
     * cancelAnimationFrame
     *
     * @description
     *
     * This function is a zone un-patched implementation of Window cancelAnimationFrame() method
     *
     * The cancelAnimationFrame() method clears a timer set with the requestAnimationFrame() method.
     * The ID value returned by requestAnimationFrame() is used as the parameter for the cancelAnimationFrame() method.
     *
     * To be able to use the cancelAnimationFrame() method, you must use a variable when creating the requestAnimationFrame method:
     *
     * const id = requestAnimationFrame("javascript function");
     * Then you will be able to stop the execution by calling the cancelAnimationFrame() method.
     *
     * cancelAnimationFrame(id);
     *
     * @param id {number} - Required. The ID value of the timer returned by the requestAnimationFrame() method
     *
     */
    function cancelAnimationFrame$1(id) {
        getZoneUnPatchedApi('cancelAnimationFrame')(id);
    }
    /**
     * setInterval
     *
     * @description
     *
     * This function is a zone un-patched implementation of Window setInterval() method
     *
     * The setInterval() method calls a function or evaluates an expression at specified intervals (in milliseconds).
     * The setInterval() method will continue calling the function until clearInterval() is called, or the window is closed.
     * The ID value returned by setInterval() is used as the parameter for the clearInterval() method.
     *
     * setInterval(cb, ms);
     *
     * @param cb - Required. The function that will be executed
     * @param ms - Required. The intervals (in milliseconds) on how often to execute the code. If the value is less than 10, the value 10 is used
     *
     */
    function setInterval(cb, ms) {
        if (ms === void 0) { ms = 0; }
        return getZoneUnPatchedApi('setInterval')(cb, ms);
    }
    /**
     * clearInterval
     *
     * @description
     *
     * This function is a zone un-patched implementation of Window clearInterval() method
     *
     * The clearInterval() method clears a timer set with the setInterval() method.
     * The ID value returned by setInterval() is used as the parameter for the clearInterval() method.
     *
     * To be able to use the clearInterval() method, you must use a variable when creating the interval method:
     *
     * const id = setInterval("javascript function", milliseconds);
     * Then you will be able to stop the execution by calling the clearInterval() method.
     *
     * clearInterval(id);
     *
     * @param id {number} - Required. The ID value of the timer returned by the setInterval() method
     *
     */
    function clearInterval(id) {
        return getZoneUnPatchedApi('clearInterval')(id);
    }
    /**
     * setTimeout
     *
     * @description
     *
     * This function is a zone un-patched implementation of Window setTimeout() method
     *
     * The setTimeout() method calls a function or evaluates an expression after a specified number of milliseconds.
     * The function is only executed once. If you need to repeat execution, use the setInterval() method.
     * Use the clearTimeout() method to prevent the function from running.
     *
     * setTimeout(cb, ms);
     *
     * @param cb - Required. The function that will be executed
     * @param ms - Optional. The number of milliseconds to wait before executing the code. If omitted, the value 0 is used
     *
     */
    function setTimeout$1(cb, ms) {
        if (ms === void 0) { ms = 0; }
        return getZoneUnPatchedApi('setTimeout')(cb, ms);
    }
    /**
     * clearTimeout
     *
     * @description
     *
     * This function is a zone un-patched implementation of Window#clearTimeout() method
     *
     * The clearTimeout() method clears a timer set with the setTimeout() method.
     * The ID value returned by setTimeout() is used as the parameter for the clearTimeout() method.
     *
     * const id = setTimeout("javascript function", milliseconds);
     * Then, if the function has not already been executed, you will be able to stop the execution by calling the clearTimeout() method.
     *
     * clearTimeout(id);
     *
     * @param id {number} -	Required. The ID value of the timer returned by the setTimeout() method
     *
     */
    function clearTimeout$1(id) {
        getZoneUnPatchedApi('clearTimeout')(id);
    }
    /**
     * This function is a zone un-patched implementation of Element#addEventListener() method.
     * @param elem
     */
    function unpatchAddEventListener(elem) {
        elem.addEventListener = getZoneUnPatchedApi('addEventListener', elem).bind(elem);
        return elem;
    }

    var animationFrameTick = function () { return new rxjs.Observable(function (subscriber) {
        var id = requestAnimationFrame$1(function () {
            subscriber.next(0);
            subscriber.complete();
        });
        return function () {
            cancelAnimationFrame$1(id);
        };
    }); };
    var ɵ0$4 = animationFrameTick;
    var ɵ1$1 = function (cdRef, _, notification) {
        cdRef.detectChanges();
    }, ɵ2$1 = function (work, scope) { return function (o$) { return o$.pipe(coalesceWith(animationFrameTick(), scope), operators.tap(function () { return work(); })); }; };
    var localCredentials = {
        name: 'local',
        work: ɵ1$1,
        behavior: ɵ2$1,
    };
    var ɵ3$1 = function (_, context) { return i0.ɵmarkDirty(context); }, ɵ4$1 = function (work) { return function (o$) { return o$.pipe(operators.tap(function () { return work(); })); }; };
    var globalCredentials = {
        name: 'global',
        work: ɵ3$1,
        behavior: ɵ4$1,
    };
    var ɵ5$1 = function () { return void 0; }, ɵ6$1 = function () { return function (o$) { return o$; }; };
    var noopCredentials = {
        name: 'noop',
        work: ɵ5$1,
        behavior: ɵ6$1,
    };
    var ɵ7$1 = function (cdRef) { return cdRef.markForCheck(); }, ɵ8$1 = function (work) { return function (o$) { return o$.pipe(operators.tap(function () { return work(); })); }; };
    var nativeCredentials = {
        name: 'native',
        work: ɵ7$1,
        behavior: ɵ8$1,
    };
    var RX_NATIVE_STRATEGIES = {
        global: globalCredentials,
        native: nativeCredentials,
        noop: noopCredentials,
        local: localCredentials,
    };

    var RX_ANGULAR_CONFIG = new i0.InjectionToken('rx-angular-config');
    var RX_ANGULAR_DEFAULTS = {
        primaryStrategy: 'normal',
        customStrategies: Object.assign(Object.assign(Object.assign({}, RX_NATIVE_STRATEGIES), { local: Object.assign(Object.assign({}, RX_CONCURRENT_STRATEGIES.immediate), { name: 'local' }) }), RX_CONCURRENT_STRATEGIES),
        patchZone: true,
    };
    function mergeDefaultConfig(cfg) {
        var custom = cfg
            ? cfg
            : {
                customStrategies: {},
            };
        return Object.assign(Object.assign(Object.assign({}, RX_ANGULAR_DEFAULTS), custom), { customStrategies: Object.assign(Object.assign({}, custom.customStrategies), RX_ANGULAR_DEFAULTS.customStrategies) });
    }

    /**
     *
     *
     * @docsCategory RenderStrategies
     * @docsPage RenderStrategies
     * @publicApi
     */
    var RxStrategyProvider = /** @class */ (function () {
        function RxStrategyProvider(cfg) {
            this._strategies$ = new rxjs.BehaviorSubject(undefined);
            this._primaryStrategy$ = new rxjs.BehaviorSubject(undefined);
            this.primaryStrategy$ = this._primaryStrategy$.asObservable();
            this.strategies$ = this._strategies$.asObservable();
            this.strategyNames$ = this.strategies$.pipe(operators.map(function (strategies) { return Object.values(strategies).map(function (s) { return s.name; }); }), operators.shareReplay({ bufferSize: 1, refCount: true }));
            this._cfg = mergeDefaultConfig(cfg);
            this._strategies$.next(this._cfg.customStrategies);
            this.primaryStrategy = this.config.primaryStrategy;
        }
        Object.defineProperty(RxStrategyProvider.prototype, "config", {
            get: function () {
                return this._cfg;
            },
            enumerable: false,
            configurable: true
        });
        Object.defineProperty(RxStrategyProvider.prototype, "strategies", {
            get: function () {
                return this._strategies$.getValue();
            },
            enumerable: false,
            configurable: true
        });
        Object.defineProperty(RxStrategyProvider.prototype, "strategyNames", {
            get: function () {
                return Object.values(this.strategies).map(function (s) { return s.name; });
            },
            enumerable: false,
            configurable: true
        });
        Object.defineProperty(RxStrategyProvider.prototype, "primaryStrategy", {
            get: function () {
                return this._primaryStrategy$.getValue().name;
            },
            set: function (strategyName) {
                this._primaryStrategy$.next(this.strategies[strategyName]);
            },
            enumerable: false,
            configurable: true
        });
        RxStrategyProvider.prototype.scheduleWith = function (work, options) {
            var strategy = this.strategies[(options === null || options === void 0 ? void 0 : options.strategy) || this.primaryStrategy];
            var scope = (options === null || options === void 0 ? void 0 : options.scope) || {};
            var _work = getWork(work, options === null || options === void 0 ? void 0 : options.patchZone);
            return function (o$) { return o$.pipe(operators.switchMap(function (v) { return onStrategy(v, strategy, function (_v) {
                _work(_v);
            }, { scope: scope }); })); };
        };
        RxStrategyProvider.prototype.schedule = function (work, options) {
            var strategy = this.strategies[(options === null || options === void 0 ? void 0 : options.strategy) || this.primaryStrategy];
            var scope = (options === null || options === void 0 ? void 0 : options.scope) || {};
            var _work = getWork(work, options === null || options === void 0 ? void 0 : options.patchZone);
            var returnVal;
            return onStrategy(null, strategy, function () {
                returnVal = _work();
            }, { scope: scope }).pipe(operators.map(function () { return returnVal; }));
        };
        RxStrategyProvider.prototype.scheduleCD = function (cdRef, options) {
            var strategy = this.strategies[(options === null || options === void 0 ? void 0 : options.strategy) || this.primaryStrategy];
            var scope = (options === null || options === void 0 ? void 0 : options.scope) || cdRef;
            var abC = (options === null || options === void 0 ? void 0 : options.abortCtrl) || new AbortController();
            var work = getWork(function () {
                strategy.work(cdRef, scope);
                if (options === null || options === void 0 ? void 0 : options.afterCD) {
                    options.afterCD();
                }
            }, options.patchZone);
            onStrategy(null, strategy, function () {
                work();
            }, { scope: scope })
                .pipe(operators.takeUntil(rxjs.fromEvent(abC.signal, 'abort')))
                .subscribe();
            return abC;
        };
        return RxStrategyProvider;
    }());
    /** @nocollapse */ RxStrategyProvider.ɵprov = i0.ɵɵdefineInjectable({ factory: function RxStrategyProvider_Factory() { return new RxStrategyProvider(i0.ɵɵinject(RX_ANGULAR_CONFIG, 8)); }, token: RxStrategyProvider, providedIn: "root" });
    RxStrategyProvider.decorators = [
        { type: i0.Injectable, args: [{ providedIn: 'root' },] }
    ];
    /** @nocollapse */
    RxStrategyProvider.ctorParameters = function () { return [
        { type: undefined, decorators: [{ type: i0.Optional }, { type: i0.Inject, args: [RX_ANGULAR_CONFIG,] }] }
    ]; };
    function getWork(work, patchZone) {
        var _work = work;
        if (patchZone) {
            _work = function (args) { return patchZone.run(function () { return work(args); }); };
        }
        return _work;
    }

    /**
     * A unit of work to be executed in a `scheduler`. An action is typically
     * created from within a {@link SchedulerLike} and an RxJS user does not need to concern
     * themselves about creating and manipulating an Action.
     *
     * ```ts
     * class Action<T> extends Subscription {
     *   new (scheduler: Scheduler, work: (state?: T) => void);
     *   schedule(state?: T, delay: number = 0): Subscription;
     * }
     * ```
     *
     * @class Action<T>
     */
    var Action = /** @class */ (function (_super) {
        __extends(Action, _super);
        function Action(scheduler, work) {
            return _super.call(this) || this;
        }
        /**
         * Schedules this action on its parent {@link SchedulerLike} for execution. May be passed
         * some context object, `state`. May happen at some point in the future,
         * according to the `delay` parameter, if specified.
         * @param {T} [state] Some contextual data that the `work` function uses when
         * called by the Scheduler.
         * @param {number} [delay] Time to wait before executing the work, where the
         * time unit is implicit and defined by the Scheduler.
         * @return {void}
         */
        Action.prototype.schedule = function (state, delay) {
            if (delay === void 0) { delay = 0; }
            return this;
        };
        return Action;
    }(rxjs.Subscription));

    /**
     * We need this JSDoc comment for affecting ESDoc.
     * @ignore
     * @extends {Ignored}
     */
    var AsyncAction = /** @class */ (function (_super) {
        __extends(AsyncAction, _super);
        function AsyncAction(scheduler, work) {
            var _this = _super.call(this, scheduler, work) || this;
            _this.scheduler = scheduler;
            _this.work = work;
            _this.pending = false;
            return _this;
        }
        AsyncAction.prototype.schedule = function (state, delay) {
            if (delay === void 0) { delay = 0; }
            if (this.closed) {
                return this;
            }
            // Always replace the current state with the new state.
            this.state = state;
            var id = this.id;
            var scheduler = this.scheduler;
            //
            // Important implementation note:
            //
            // Actions only execute once by default, unless rescheduled from within the
            // scheduled callback. This allows us to implement single and repeat
            // actions via the same code path, without adding API surface area, as well
            // as mimic traditional recursion but across asynchronous boundaries.
            //
            // However, JS runtimes and timers distinguish between intervals achieved by
            // serial `setTimeout` calls vs. a single `setInterval` call. An interval of
            // serial `setTimeout` calls can be individually delayed, which delays
            // scheduling the next `setTimeout`, and so on. `setInterval` attempts to
            // guarantee the interval callback will be invoked more precisely to the
            // interval period, regardless of load.
            //
            // Therefore, we use `setInterval` to schedule single and repeat actions.
            // If the action reschedules itself with the same delay, the interval is not
            // canceled. If the action doesn't reschedule, or reschedules with a
            // different delay, the interval will be canceled after scheduled callback
            // execution.
            //
            if (id != null) {
                this.id = this.recycleAsyncId(scheduler, id, delay);
            }
            // Set the pending flag indicating that this action has been scheduled, or
            // has recursively rescheduled itself.
            this.pending = true;
            this.delay = delay;
            // If this action has already an async Id, don't request a new one.
            this.id = this.id || this.requestAsyncId(scheduler, this.id, delay);
            return this;
        };
        AsyncAction.prototype.requestAsyncId = function (scheduler, id, delay) {
            if (delay === void 0) { delay = 0; }
            return setInterval(scheduler.flush.bind(scheduler, this), delay);
        };
        AsyncAction.prototype.recycleAsyncId = function (scheduler, id, delay) {
            if (delay === void 0) { delay = 0; }
            // If this action is rescheduled with the same delay time, don't clear the interval id.
            if (delay !== null && this.delay === delay && this.pending === false) {
                return id;
            }
            // Otherwise, if the action's delay time is different from the current delay,
            // or the action has been rescheduled before it's executed, clear the interval id
            clearInterval(id);
            return undefined;
        };
        /**
         * Immediately executes this action and the `work` it contains.
         * @return {any}
         */
        AsyncAction.prototype.execute = function (state, delay) {
            if (this.closed) {
                return new Error('executing a cancelled action');
            }
            this.pending = false;
            var error = this._execute(state, delay);
            if (error) {
                return error;
            }
            else if (this.pending === false && this.id != null) {
                // Dequeue if the action didn't reschedule itself. Don't call
                // unsubscribe(), because the action could reschedule later.
                // For example:
                // ```
                // scheduler.schedule(function doWork(counter) {
                //   /* ... I'm a busy worker bee ... */
                //   var originalAction = this;
                //   /* wait 100ms before rescheduling the action */
                //   setTimeout(function () {
                //     originalAction.schedule(counter + 1);
                //   }, 100);
                // }, 1000);
                // ```
                this.id = this.recycleAsyncId(this.scheduler, this.id, null);
            }
        };
        AsyncAction.prototype._execute = function (state, delay) {
            var errored = false;
            var errorValue;
            try {
                this.work(state);
            }
            catch (e) {
                errored = true;
                errorValue = (!!e && e) || new Error(e);
            }
            if (errored) {
                this.unsubscribe();
                return errorValue;
            }
        };
        /** @deprecated This is an internal implementation detail, do not use. */
        AsyncAction.prototype._unsubscribe = function () {
            var id = this.id;
            var scheduler = this.scheduler;
            var actions = scheduler.actions;
            // @ts-ignore
            var index = actions.indexOf(this);
            this.work = null;
            this.state = null;
            this.pending = false;
            this.scheduler = null;
            if (index !== -1) {
                actions.splice(index, 1);
            }
            if (id != null) {
                this.id = this.recycleAsyncId(scheduler, id, null);
            }
            this.delay = null;
        };
        return AsyncAction;
    }(Action));

    function nowFn() {
        var fn = function () { return Date.now(); };
        return fn;
    }
    /**
     * An execution context and a data structure to order tasks and schedule their
     * execution. Provides a notion of (potentially virtual) time, through the
     * `now()` getter method.
     *
     * Each unit of work in a Scheduler is called an `Action`.
     *
     * ```ts
     * class Scheduler {
     *   now(): number;
     *   schedule(work, delay?, state?): Subscription;
     * }
     * ```
     *
     * @class Scheduler
     * @deprecated Scheduler is an internal implementation detail of RxJS, and
     * should not be used directly. Rather, create your own class and implement
     * {@link SchedulerLike}
     */
    var Scheduler = /** @class */ (function () {
        function Scheduler(SchedulerAction, now) {
            if (now === void 0) { now = Scheduler.now; }
            this.SchedulerAction = SchedulerAction;
            this.now = now;
        }
        /**
         * Schedules a function, `work`, for execution. May happen at some point in
         * the future, according to the `delay` parameter, if specified. May be passed
         * some context object, `state`, which will be passed to the `work` function.
         *
         * The given arguments will be processed an stored as an Action object in a
         * queue of actions.
         *
         * @param {function(state: ?T): ?Subscription} work A function representing a
         * task, or some unit of work to be executed by the Scheduler.
         * @param {number} [delay] Time to wait before executing the work, where the
         * time unit is implicit and defined by the Scheduler itself.
         * @param {T} [state] Some contextual data that the `work` function uses when
         * called by the Scheduler.
         * @return {Subscription} A subscription in order to be able to unsubscribe
         * the scheduled work.
         */
        Scheduler.prototype.schedule = function (work, delay, state) {
            if (delay === void 0) { delay = 0; }
            return new this.SchedulerAction(this, work).schedule(state, delay);
        };
        return Scheduler;
    }());
    /**
     * Note: the extra arrow function wrapper is to make testing by overriding
     * Date.now easier.
     * @nocollapse
     */
    Scheduler.now = nowFn();

    var AsyncScheduler = /** @class */ (function (_super) {
        __extends(AsyncScheduler, _super);
        function AsyncScheduler(SchedulerAction, now) {
            if (now === void 0) { now = Scheduler.now; }
            var _this = _super.call(this, SchedulerAction, function () {
                if (AsyncScheduler.delegate && AsyncScheduler.delegate !== _this) {
                    return AsyncScheduler.delegate.now();
                }
                else {
                    return now();
                }
            }) || this;
            _this.actions = [];
            /**
             * A flag to indicate whether the Scheduler is currently executing a batch of
             * queued actions.
             * @type {boolean}
             * @deprecated internal use only
             */
            _this.active = false;
            /**
             * An internal ID used to track the latest asynchronous task such as those
             * coming from `setTimeout`, `setInterval`, `requestAnimationFrame`, and
             * others.
             * @type {any}
             * @deprecated internal use only
             */
            _this.scheduled = undefined;
            return _this;
        }
        AsyncScheduler.prototype.schedule = function (work, delay, state) {
            if (delay === void 0) { delay = 0; }
            if (AsyncScheduler.delegate && AsyncScheduler.delegate !== this) {
                return AsyncScheduler.delegate.schedule(work, delay, state);
            }
            else {
                return _super.prototype.schedule.call(this, work, delay, state);
            }
        };
        AsyncScheduler.prototype.flush = function (action) {
            var actions = this.actions;
            if (this.active) {
                actions.push(action);
                return;
            }
            var error;
            this.active = true;
            do {
                if ((error = action.execute(action.state, action.delay))) {
                    break;
                }
                // @ts-ignore
            } while ((action = actions.shift())); // exhaust the scheduler queue
            this.active = false;
            if (error) {
                // @ts-ignore
                while ((action = actions.shift())) {
                    action.unsubscribe();
                }
                throw error;
            }
        };
        return AsyncScheduler;
    }(Scheduler));
    AsyncScheduler.delegate = null;

    // tslint:disable file-name-casing
    /**
     *
     * NOTE: This is a zone un-patched version of rxjs asyncScheduler
     *
     * Async Scheduler
     *
     * <span class="informal">Schedule task as if you used setTimeout(task, duration)</span>
     *
     * `async` scheduler schedules tasks asynchronously, by putting them on the JavaScript
     * event loop queue. It is best used to delay tasks in time or to schedule tasks repeating
     * in intervals.
     *
     * If you just want to "defer" task, that is to perform it right after currently
     * executing synchronous code ends (commonly achieved by `setTimeout(deferredTask, 0)`),
     * better choice will be the {@link asapScheduler} scheduler.
     *
     * ## Examples
     * Use async scheduler to delay task
     * ```ts
     * import { asyncScheduler } from '@cu/perf-utils';
     *
     * const task = () => console.log('it works!');
     *
     * asyncScheduler.schedule(task, 2000);
     *
     * // After 2 seconds logs:
     * // "it works!"
     * ```
     *
     * Use async scheduler to repeat task in intervals
     * ```ts
     * import { asyncScheduler } from '@cu/perf-utils';
     *
     * function task(state) {
     *   console.log(state);
     *   this.schedule(state + 1, 1000); // `this` references currently executing Action,
     *                                   // which we reschedule with new state and delay
     * }
     *
     * asyncScheduler.schedule(task, 3000, 0);
     *
     * // Logs:
     * // 0 after 3s
     * // 1 after 4s
     * // 2 after 5s
     * // 3 after 6s
     * ```
     */
    var async = new AsyncScheduler(AsyncAction);

    /**
     * We need this JSDoc comment for affecting ESDoc.
     * @ignore
     * @extends {Ignored}
     */
    var AsapAction = /** @class */ (function (_super) {
        __extends(AsapAction, _super);
        function AsapAction(scheduler, work) {
            var _this = _super.call(this, scheduler, work) || this;
            _this.scheduler = scheduler;
            _this.work = work;
            return _this;
        }
        AsapAction.prototype.requestAsyncId = function (scheduler, id, delay) {
            if (delay === void 0) { delay = 0; }
            // If delay is greater than 0, request as an async action.
            if (delay !== null && delay > 0) {
                return _super.prototype.requestAsyncId.call(this, scheduler, id, delay);
            }
            // Push the action to the end of the scheduler queue.
            scheduler.actions.push(this);
            // If a microtask has already been scheduled, don't schedule another
            // one. If a microtask hasn't been scheduled yet, schedule one now. Return
            // the current scheduled microtask id.
            return (scheduler.scheduled ||
                (scheduler.scheduled = Immediate.setImmediate(scheduler.flush.bind(scheduler, undefined))));
        };
        AsapAction.prototype.recycleAsyncId = function (scheduler, id, delay) {
            if (delay === void 0) { delay = 0; }
            // If delay exists and is greater than 0, or if the delay is null (the
            // action wasn't rescheduled) but was originally scheduled as an async
            // action, then recycle as an async action.
            if ((delay !== null && delay > 0) || (delay === null && this.delay > 0)) {
                return _super.prototype.recycleAsyncId.call(this, scheduler, id, delay);
            }
            // If the scheduler queue is empty, cancel the requested microtask and
            // set the scheduled flag to undefined so the next AsapAction will schedule
            // its own.
            if (scheduler.actions.length === 0) {
                Immediate.clearImmediate(id);
                scheduler.scheduled = undefined;
            }
            // Return undefined so the action knows to request a new async id if it's rescheduled.
            return undefined;
        };
        return AsapAction;
    }(AsyncAction));
    var nextHandle = 1;
    // The promise needs to be created lazily otherwise it won't be patched by Zones
    var resolved;
    var activeHandles = {};
    /**
     * Finds the handle in the list of active handles, and removes it.
     * Returns `true` if found, `false` otherwise. Used both to clear
     * Immediate scheduled tasks, and to identify if a task should be scheduled.
     */
    function findAndClearHandle(handle) {
        if (handle in activeHandles) {
            delete activeHandles[handle];
            return true;
        }
        return false;
    }
    /**
     * Helper functions to schedule and unschedule microtasks.
     */
    var Immediate = {
        setImmediate: function (cb) {
            var handle = nextHandle++;
            activeHandles[handle] = true;
            if (!resolved) {
                resolved = Promise$1.resolve();
            }
            resolved.then(function () { return findAndClearHandle(handle) && cb(); });
            return handle;
        },
        clearImmediate: function (handle) {
            findAndClearHandle(handle);
        },
    };
    /**
     * Used for internal testing purposes only. Do not export from library.
     */
    var TestTools = {
        pending: function () {
            return Object.keys(activeHandles).length;
        },
    };

    var AsapScheduler = /** @class */ (function (_super) {
        __extends(AsapScheduler, _super);
        function AsapScheduler() {
            return _super !== null && _super.apply(this, arguments) || this;
        }
        AsapScheduler.prototype.flush = function (action) {
            this.active = true;
            this.scheduled = undefined;
            var actions = this.actions;
            var error;
            var index = -1;
            var count = actions.length;
            action = action || actions.shift();
            do {
                if ((error = action.execute(action.state, action.delay))) {
                    break;
                }
            } while (++index < count && (action = actions.shift()));
            this.active = false;
            if (error) {
                while (++index < count && (action = actions.shift())) {
                    action.unsubscribe();
                }
                throw error;
            }
        };
        return AsapScheduler;
    }(AsyncScheduler));

    // tslint:disable file-name-casing
    /**
     *
     * NOTE: This is a zone un-patched version of rxjs asapScheduler
     *
     * Asap Scheduler
     *
     * <span class="informal">Perform task as fast as it can be performed asynchronously</span>
     *
     * `asap` scheduler behaves the same as {@link asyncScheduler} scheduler when you use it to delay task
     * in time. If however you set delay to `0`, `asap` will wait for current synchronously executing
     * code to end and then it will try to execute given task as fast as possible.
     *
     * `asap` scheduler will do its best to minimize time between end of currently executing code
     * and start of scheduled task. This makes it best candidate for performing so called "deferring".
     * Traditionally this was achieved by calling `setTimeout(deferredTask, 0)`, but that technique involves
     * some (although minimal) unwanted delay.
     *
     * Note that using `asap` scheduler does not necessarily mean that your task will be first to process
     * after currently executing code. In particular, if some task was also scheduled with `asap` before,
     * that task will execute first. That being said, if you need to schedule task asynchronously, but
     * as soon as possible, `asap` scheduler is your best bet.
     *
     * ## Example
     * Compare async and asap scheduler<
     * ```ts
     * import { asapScheduler, asyncScheduler } from '@cu/perf-utils';
     *
     * asyncScheduler.schedule(() => console.log('async')); // scheduling 'async' first...
     * asapScheduler.schedule(() => console.log('asap'));
     *
     * // Logs:
     * // "asap"
     * // "async"
     * // ... but 'asap' goes first!
     * ```
     */
    var asap = new AsapScheduler(AsapAction);

    /**
     * We need this JSDoc comment for affecting ESDoc.
     * @ignore
     * @extends {Ignored}
     */
    var QueueAction = /** @class */ (function (_super) {
        __extends(QueueAction, _super);
        function QueueAction(scheduler, work) {
            var _this = _super.call(this, scheduler, work) || this;
            _this.scheduler = scheduler;
            _this.work = work;
            return _this;
        }
        QueueAction.prototype.schedule = function (state, delay) {
            if (delay === void 0) { delay = 0; }
            if (delay > 0) {
                return _super.prototype.schedule.call(this, state, delay);
            }
            this.delay = delay;
            this.state = state;
            // @ts-ignore
            this.scheduler.flush(this);
            return this;
        };
        QueueAction.prototype.execute = function (state, delay) {
            return delay > 0 || this.closed
                ? _super.prototype.execute.call(this, state, delay)
                : this._execute(state, delay);
        };
        QueueAction.prototype.requestAsyncId = function (scheduler, id, delay) {
            if (delay === void 0) { delay = 0; }
            // If delay exists and is greater than 0, or if the delay is null (the
            // action wasn't rescheduled) but was originally scheduled as an async
            // action, then recycle as an async action.
            if ((delay !== null && delay > 0) || (delay === null && this.delay > 0)) {
                return _super.prototype.requestAsyncId.call(this, scheduler, id, delay);
            }
            // Otherwise flush the scheduler starting with this action.
            // @ts-ignore
            return scheduler.flush(this);
        };
        return QueueAction;
    }(AsyncAction));

    var QueueScheduler = /** @class */ (function (_super) {
        __extends(QueueScheduler, _super);
        function QueueScheduler() {
            return _super !== null && _super.apply(this, arguments) || this;
        }
        return QueueScheduler;
    }(AsyncScheduler));

    // tslint:disable file-name-casing
    /**
     *
     * NOTE: This is a zone un-patched version of rxjs queueScheduler
     *
     * Queue Scheduler
     *
     * <span class="informal">Put every next task on a queue, instead of executing it immediately</span>
     *
     * `queue` scheduler, when used with delay, behaves the same as {@link asyncScheduler} scheduler.
     *
     * When used without delay, it schedules given task synchronously - executes it right when
     * it is scheduled. However when called recursively, that is when inside the scheduled task,
     * another task is scheduled with queue scheduler, instead of executing immediately as well,
     * that task will be put on a queue and wait for current one to finish.
     *
     * This means that when you execute task with `queue` scheduler, you are sure it will end
     * before any other task scheduled with that scheduler will start.
     *
     * ## Examples
     * Schedule recursively first, then do something
     * ```ts
     * import { queueScheduler } from '@cu/perf-utils';
     *
     * queueScheduler.schedule(() => {
     *   queueScheduler.schedule(() => console.log('second')); // will not happen now, but will be put on a queue
     *
     *   console.log('first');
     * });
     *
     * // Logs:
     * // "first"
     * // "second"
     * ```
     *
     * Reschedule itself recursively
     * ```ts
     * import { queueScheduler } from '@cu/perf-utils';
     *
     * queueScheduler.schedule(function(state) {
     *   if (state !== 0) {
     *     console.log('before', state);
     *     this.schedule(state - 1); // `this` references currently executing Action,
     *                               // which we reschedule with new state
     *     console.log('after', state);
     *   }
     * }, 0, 3);
     *
     * // In scheduler that runs recursively, you would expect:
     * // "before", 3
     * // "before", 2
     * // "before", 1
     * // "after", 1
     * // "after", 2
     * // "after", 3
     *
     * // But with queue it logs:
     * // "before", 3
     * // "after", 3
     * // "before", 2
     * // "after", 2
     * // "before", 1
     * // "after", 1
     * ```
     */
    var queue = new QueueScheduler(QueueAction);

    /**
     * We need this JSDoc comment for affecting ESDoc.
     * @ignore
     * @extends {Ignored}
     */
    var AnimationFrameAction = /** @class */ (function (_super) {
        __extends(AnimationFrameAction, _super);
        function AnimationFrameAction(scheduler, work) {
            var _this = _super.call(this, scheduler, work) || this;
            _this.scheduler = scheduler;
            _this.work = work;
            return _this;
        }
        AnimationFrameAction.prototype.requestAsyncId = function (scheduler, id, delay) {
            if (delay === void 0) { delay = 0; }
            // If delay is greater than 0, request as an async action.
            if (delay !== null && delay > 0) {
                return _super.prototype.requestAsyncId.call(this, scheduler, id, delay);
            }
            // Push the action to the end of the scheduler queue.
            // @ts-ignore
            scheduler.actions.push(this);
            // If an animation frame has already been requested, don't request another
            // one. If an animation frame hasn't been requested yet, request one. Return
            // the current animation frame request id.
            return (scheduler.scheduled ||
                (scheduler.scheduled = requestAnimationFrame$1(function () { return scheduler.flush(undefined); })));
        };
        AnimationFrameAction.prototype.recycleAsyncId = function (scheduler, id, delay) {
            if (delay === void 0) { delay = 0; }
            // If delay exists and is greater than 0, or if the delay is null (the
            // action wasn't rescheduled) but was originally scheduled as an async
            // action, then recycle as an async action.
            if ((delay !== null && delay > 0) || (delay === null && this.delay > 0)) {
                return _super.prototype.recycleAsyncId.call(this, scheduler, id, delay);
            }
            // If the scheduler queue is empty, cancel the requested animation frame and
            // set the scheduled flag to undefined so the next AnimationFrameAction will
            // request its own.
            if (scheduler.actions.length === 0) {
                cancelAnimationFrame$1(id);
                scheduler.scheduled = undefined;
            }
            // Return undefined so the action knows to request a new async id if it's rescheduled.
            return undefined;
        };
        return AnimationFrameAction;
    }(AsyncAction));

    var AnimationFrameScheduler = /** @class */ (function (_super) {
        __extends(AnimationFrameScheduler, _super);
        function AnimationFrameScheduler() {
            return _super !== null && _super.apply(this, arguments) || this;
        }
        AnimationFrameScheduler.prototype.flush = function (action) {
            this.active = true;
            this.scheduled = undefined;
            var actions = this.actions;
            var error;
            var index = -1;
            var count = actions.length;
            action = action || actions.shift();
            do {
                if ((error = action.execute(action.state, action.delay))) {
                    break;
                }
            } while (++index < count && (action = actions.shift()));
            this.active = false;
            if (error) {
                while (++index < count && (action = actions.shift())) {
                    action.unsubscribe();
                }
                throw error;
            }
        };
        return AnimationFrameScheduler;
    }(AsyncScheduler));

    // tslint:disable file-name-casing
    /**
     *
     * NOTE: This is a zone un-patched version of rxjs animationFrameScheduler
     *
     * Animation Frame Scheduler
     *
     * <span class="informal">Perform task when `window.requestAnimationFrame` would fire</span>
     *
     * When `animationFrame` scheduler is used with delay, it will fall back to {@link asyncScheduler} scheduler
     * behaviour.
     *
     * Without delay, `animationFrame` scheduler can be used to create smooth browser animations.
     * It makes sure scheduled task will happen just before next browser content repaint,
     * thus performing animations as efficiently as possible.
     *
     * ## Example
     * Schedule div height animation
     * ```ts
     * // html: <div style="background: #0ff;"></div>
     * import { animationFrameScheduler } from '@cu/perf-utils';
     *
     * const div = document.querySelector('div');
     *
     * animationFrameScheduler.schedule(function(height) {
     *   div.style.height = height + "px";
     *
     *   this.schedule(height + 1);  // `this` references currently executing Action,
     *                               // which we reschedule with new state
     * }, 0, 0);
     *
     * // You will see a div element growing in height
     * ```
     */
    var animationFrame = new AnimationFrameScheduler(AnimationFrameAction);

    /** @internal **/
    function isRxRenderError(e) {
        return (e != null && Array.isArray(e) && e.length === 2 && e[0] instanceof Error);
    }
    /** @internal **/
    function createErrorHandler(_handler) {
        var _handleError = _handler
            ? function (e) { return _handler.handleError(e); }
            : console.error;
        return {
            handleError: function (error) {
                if (isRxRenderError(error)) {
                    _handleError(error[0]);
                    console.error('additionalErrorContext', error[1]);
                }
                else {
                    _handleError(error);
                }
            },
        };
    }
    /** @internal **/
    function toRenderError(e, context) {
        return [e, context];
    }

    // Below are constants for LView indices to help us look up LView members
    // without having to remember the specific indices.
    // Uglify will inline these when minifying so there shouldn't be a cost.
    var TVIEW = 1;
    var T_HOST = 6;
    var L_CONTAINER_NATIVE = 7;
    var CONTEXT = 8;
    var HEADER_OFFSET = 20;
    /**
     * @internal
     *
     * Returns the TNode of the passed node form TVIEW of passed cdRef
     *
     * @param cdRef
     * @param native
     */
    function getTNode(cdRef, native) {
        var lView = cdRef._cdRefInjectingView;
        if (!lView) {
            return undefined;
        }
        var tView = lView[TVIEW];
        var i = HEADER_OFFSET;
        var lContainer;
        while (!lContainer && i <= tView['bindingStartIndex']) {
            var candidate = lView[i];
            if (candidate && candidate[L_CONTAINER_NATIVE] === native) {
                lContainer = candidate;
            }
            i++;
        }
        return lContainer[T_HOST];
    }
    /**
     * @internal
     *
     * Returns a set of references to parent views
     *
     *
     * @param cdRef
     * @param tNode
     */
    function extractProjectionParentViewSet(cdRef, tNode) {
        var injectingLView = cdRef._cdRefInjectingView;
        var injectingTView = injectingLView[1];
        var components = new Set(injectingTView['components']);
        var parentElements = new Set();
        var parent = tNode['parent'];
        while (parent != null && components.size > 0) {
            var idx = parent['index'];
            if (components.has(idx)) {
                // TODO: we should discuss about this. currently only the first Component will get returned, not a list of
                //  components. Maybe we should make the parent notification configurable regarding the level of `deepness`?
                // components.delete(idx);
                components.clear();
                parentElements.add(injectingLView[idx][CONTEXT]);
            }
            parent = parent['parent'];
        }
        return parentElements;
    }
    function extractProjectionViews(cdRef, tNode) {
        return Array.from(extractProjectionParentViewSet(cdRef, tNode));
    }
    /**
     * A side effect operator similar to `tap` but with a static logic
     *
     *
     *
     * @param cdRef
     * @param tNode
     * @param strategy$
     */
    function renderProjectionParents(cdRef, tNode, strategy$) {
        return function (o$) { return o$.pipe(operators.withLatestFrom(strategy$), operators.switchMap(function (_a) {
            var e_1, _b;
            var _c = __read(_a, 2), _ = _c[0], strategy = _c[1];
            var parentElements = extractProjectionParentViewSet(cdRef, tNode);
            var behaviors = [];
            var _loop_1 = function (el) {
                behaviors.push(onStrategy(el, strategy, function (value, work, options) {
                    i0.ɵdetectChanges(el);
                }, { scope: el }));
            };
            try {
                for (var _d = __values(parentElements.values()), _e = _d.next(); !_e.done; _e = _d.next()) {
                    var el = _e.value;
                    _loop_1(el);
                }
            }
            catch (e_1_1) { e_1 = { error: e_1_1 }; }
            finally {
                try {
                    if (_e && !_e.done && (_b = _d.return)) _b.call(_d);
                }
                finally { if (e_1) throw e_1.error; }
            }
            behaviors.push(onStrategy(null, strategy, function (value, work, options) { return work(cdRef, options.scope); }, { scope: cdRef.context || cdRef }));
            return rxjs.merge.apply(void 0, __spread(behaviors));
        })); };
    }
    /**
     * @internal
     * creates an embeddedViewRef
     *
     * @param viewContainerRef
     * @param templateRef
     * @param context
     * @param index
     * @return EmbeddedViewRef<C>
     */
    function createEmbeddedView(viewContainerRef, templateRef, context, index) {
        if (index === void 0) { index = 0; }
        var view = viewContainerRef.createEmbeddedView(templateRef, context, index);
        view.detectChanges();
        return view;
    }
    /**
     * @internal
     *
     * A factory function returning an object to handle `TemplateRef`'s.
     * You can add and get a `TemplateRef`.
     *
     */
    function templateHandling(viewContainerRef) {
        var templateCache = new Map();
        var get = function (name) {
            return templateCache.get(name);
        };
        return {
            add: function (name, templateRef) {
                assertTemplate(name, templateRef);
                if (!templateCache.has(name)) {
                    templateCache.set(name, templateRef);
                }
                else {
                    throw new Error('Updating an already existing Template is not supported at the moment.');
                }
            },
            get: get,
            createEmbeddedView: function (name, context) { return createEmbeddedView(viewContainerRef, get(name), context); },
        };
        //
        function assertTemplate(property, templateRef) {
            var isTemplateRefOrNull = !!(!templateRef || templateRef.createEmbeddedView);
            if (!isTemplateRefOrNull) {
                throw new Error(property + " must be a TemplateRef, but received something else.");
            }
            return isTemplateRefOrNull;
        }
    }
    /**
     * @internal
     *
     * A side effect operator similar to `tap` but with a static internal logic.
     * It calls detect changes on the 'VirtualParent' and the injectingViewCdRef.
     *
     * @param tNode
     * @param injectingViewCdRef
     * @param strategy
     * @param notifyNeeded
     */
    function notifyAllParentsIfNeeded(tNode, injectingViewCdRef, strategy, notifyNeeded) {
        return function (o$) { return o$.pipe(operators.delay(0, async), operators.switchMap(function (v) {
            var notifyParent = notifyNeeded();
            if (!notifyParent) {
                return rxjs.of(v);
            }
            var behaviors = tNode
                ? getVirtualParentNotifications$(tNode, injectingViewCdRef, strategy)
                : [];
            // @TODO remove this CD on injectingViewCdRef if possible
            behaviors.push(onStrategy(null, strategy, function (_v, work, options) { return work(injectingViewCdRef, options.scope); }, { scope: injectingViewCdRef.context || injectingViewCdRef }, function (error) { return toRenderError(error, injectingViewCdRef); }));
            if (behaviors.length === 1) {
                return rxjs.of(v);
            }
            return rxjs.concat(rxjs.of(v), rxjs.combineLatest(behaviors).pipe(operators.ignoreElements()));
        })); };
    }
    /**
     * @internal
     *
     * returns an Observable executing a side effects for change detection of parents
     *
     * @param injectingViewCdRef
     * @param strategy
     * @param notify
     */
    function notifyInjectingParentIfNeeded(injectingViewCdRef, strategy, notify) {
        return rxjs.concat(rxjs.of(null), notify
            ? onStrategy(null, strategy, function (value, work, options) {
                // console.log('notify injectingView', injectingViewCdRef);
                work(injectingViewCdRef, options.scope);
            }, {}, function (error) { return toRenderError(error, injectingViewCdRef); }).pipe(operators.ignoreElements())
            : []);
    }
    /**
     * @internal
     *
     * Returns an array of observables triggering `detectChanges` on the __virtual parent__  (parent of the projected view)
     *
     * @param tNode - is a component that was projected into another component (virtual parent)
     * @param injectingViewCdRef - is needed to get the
     * @param strategy - the strategy to run the change detection
     */
    function getVirtualParentNotifications$(tNode, injectingViewCdRef, strategy) {
        var e_2, _a;
        var parentElements = extractProjectionParentViewSet(injectingViewCdRef, tNode);
        var behaviors = [];
        var _loop_2 = function (parentComponent) {
            behaviors.push(onStrategy(parentComponent, strategy, 
            // Here we CD the parent to update their projected views scenarios
            function (value, work, options) {
                // console.log('parentComponent', parentComponent);
                i0.ɵdetectChanges(parentComponent);
            }, { scope: parentComponent }, function (error) { return toRenderError(error, injectingViewCdRef); }));
        };
        try {
            for (var _b = __values(parentElements.values()), _c = _b.next(); !_c.done; _c = _b.next()) {
                var parentComponent = _c.value;
                _loop_2(parentComponent);
            }
        }
        catch (e_2_1) { e_2 = { error: e_2_1 }; }
        finally {
            try {
                if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
            }
            finally { if (e_2) throw e_2.error; }
        }
        return behaviors;
    }

    exports.RxBaseTemplateNames = void 0;
    (function (RxBaseTemplateNames) {
        RxBaseTemplateNames["error"] = "errorTpl";
        RxBaseTemplateNames["complete"] = "completeTpl";
        RxBaseTemplateNames["suspense"] = "suspenseTpl";
    })(exports.RxBaseTemplateNames || (exports.RxBaseTemplateNames = {}));

    /**
     * @internal
     *
     * A factory function that returns a map of projections to turn a notification of a Observable (next, error, complete)
     *
     * @param customNextContext - projection function to provide custom properties as well as override existing
     */
    function notificationKindToViewContext(customNextContext) {
        // @TODO rethink overrides
        return {
            suspense: function (notification) {
                var $implicit = notification.value;
                return Object.assign({ $implicit: $implicit, $suspense: true, $error: false, $complete: false }, customNextContext($implicit));
            },
            next: function (notification) {
                var $implicit = notification.value;
                return Object.assign({ $implicit: $implicit, $suspense: false, $error: false, $complete: false }, customNextContext($implicit));
            },
            error: function (notification) {
                var $implicit = notification.value;
                return Object.assign({ $implicit: $implicit, $complete: false, $error: notification.error, $suspense: false }, customNextContext($implicit));
            },
            complete: function (notification) {
                var $implicit = notification.value;
                return Object.assign({ $implicit: $implicit, $error: false, $complete: true, $suspense: false }, customNextContext($implicit));
            },
        };
    }
    function createTemplateManager(config) {
        var renderSettings = config.renderSettings, notificationToTemplateName = config.notificationToTemplateName, templateSettings = config.templateSettings;
        var defaultStrategyName = renderSettings.defaultStrategyName, strategies = renderSettings.strategies, injectingViewCdRef = renderSettings.cdRef, patchZone = renderSettings.patchZone, parent = renderSettings.parent, eRef = renderSettings.eRef;
        var errorHandler = createErrorHandler(renderSettings.errorHandler);
        var tNode = parent
            ? getTNode(injectingViewCdRef, eRef.nativeElement)
            : false;
        var activeTemplate;
        var strategyHandling$ = strategyHandling(defaultStrategyName, strategies);
        var templates = templateHandling(templateSettings.viewContainerRef);
        var viewContainerRef = templateSettings.viewContainerRef;
        var triggerHandling = config.templateTrigger$ || rxjs.EMPTY;
        var getContext = notificationKindToViewContext(templateSettings.customContext || (function (v) { }));
        var workFactory = patchZone
            ? function (work) { return patchZone.run(work); }
            : function (work) { return work(); };
        return {
            addTemplateRef: templates.add,
            // addTrigger: triggerHandling.next,
            nextStrategy: strategyHandling$.next,
            render: function (values$) {
                return values$.pipe(
                /* tslint:disable */
                // mergeWith(triggerHandling.trigger$ || EMPTY),
                /* tslint:enable */
                operators.withLatestFrom(strategyHandling$.strategy$), 
                // Cancel old renders
                operators.switchMap(function (_a) {
                    var _b = __read(_a, 2), notification = _b[0], strategy = _b[1];
                    var kind = notification.kind;
                    var value = notification.value;
                    var templateName = notificationToTemplateName[kind](value, templates);
                    var template = templates.get(templateName);
                    var isNewTemplate = activeTemplate !== templateName;
                    var notifyParent = isNewTemplate && parent;
                    return rxjs.merge(onStrategy(value, strategy, function (v, work, options) {
                        var context = getContext[kind](notification);
                        if (isNewTemplate) {
                            // template has changed (undefined => next; suspense => next; ...)
                            // handle remove & insert
                            // remove current view if there is any
                            if (viewContainerRef.length > 0) {
                                // patch removal if needed
                                workFactory(function () { return viewContainerRef.clear(); });
                            }
                            // create new view if any
                            if (template) {
                                // createEmbeddedView is already patched, no need for workFactory
                                workFactory(function () { return templates.createEmbeddedView(templateName, context); });
                            }
                        }
                        else if (template) {
                            // template didn't change, update it
                            // handle update
                            var view_1 = viewContainerRef.get(0);
                            Object.keys(context).forEach(function (k) {
                                view_1.context[k] = context[k];
                            });
                            // update view context, patch if needed
                            workFactory(function () { return work(view_1, options.scope, notification); });
                        }
                        activeTemplate = templateName;
                    }
                    // we don't need to specify any scope here. The template manager is the only one
                    // who will call `viewRef#detectChanges` on any of the templates it manages.
                    // whenever a new value comes in, any pre-scheduled work of this taskManager will
                    // be nooped before a new work will be scheduled. This happens because of the implementation
                    // of `StrategyCredential#behavior`
                    ).pipe(notifyAllParentsIfNeeded(tNode, injectingViewCdRef, strategy, function () { return notifyParent; })), notifyInjectingParentIfNeeded(injectingViewCdRef, strategy, isNewTemplate).pipe(operators.ignoreElements())).pipe(operators.catchError(function (e) {
                        errorHandler.handleError(e);
                        return rxjs.of(e);
                    }));
                }));
            },
        };
    }

    /**
     * @internal
     *
     * Factory that returns a `ListTemplateManager` for the passed params.
     *
     * @param templateSettings
     */
    function getTemplateHandler(templateSettings) {
        var viewContainerRef = templateSettings.viewContainerRef, initialTemplateRef = templateSettings.initialTemplateRef, createViewContext = templateSettings.createViewContext, updateViewContext = templateSettings.updateViewContext, patchZone = templateSettings.patchZone;
        var workFactory = patchZone
            ? function (work) { return patchZone.run(work); }
            : function (work) { return work(); };
        return {
            updateUnchangedContext: updateUnchangedContext,
            insertView: insertView,
            moveView: moveView,
            removeView: removeView,
            getListChanges: getListChanges,
            updateView: updateView,
        };
        // =====
        function updateUnchangedContext(index, count) {
            var view = viewContainerRef.get(index);
            workFactory(function () {
                view.context.updateContext({
                    count: count,
                    index: index,
                });
                view.detectChanges();
            });
        }
        function moveView(oldIndex, item, index, count) {
            var oldView = viewContainerRef.get(oldIndex);
            var view = viewContainerRef.move(oldView, index);
            workFactory(function () {
                updateViewContext(item, view, {
                    count: count,
                    index: index,
                });
                view.detectChanges();
            });
        }
        function updateView(item, index, count) {
            var view = viewContainerRef.get(index);
            workFactory(function () {
                updateViewContext(item, view, {
                    count: count,
                    index: index,
                });
                view.detectChanges();
            });
        }
        function removeView(index) {
            return workFactory(function () { return viewContainerRef.remove(index); });
        }
        function insertView(item, index, count) {
            workFactory(function () {
                createEmbeddedView(viewContainerRef, initialTemplateRef, createViewContext(item, {
                    count: count,
                    index: index,
                }), index);
            });
        }
    }
    /**
     * @internal
     *
     * @param changes
     * @param items
     */
    function getListChanges(changes, items) {
        var changedIdxs = new Set();
        var changesArr = [];
        var notifyParent = false;
        changes.forEachOperation(function (record, adjustedPreviousIndex, currentIndex) {
            var item = record.item;
            if (record.previousIndex == null) {
                // insert
                changesArr.push(getInsertChange(item, currentIndex));
                changedIdxs.add(item);
                notifyParent = true;
            }
            else if (currentIndex == null) {
                // remove
                changesArr.push(getRemoveChange(item, adjustedPreviousIndex));
                changedIdxs.add(item);
                notifyParent = true;
            }
            else if (adjustedPreviousIndex !== null) {
                // move
                changesArr.push(getMoveChange(item, currentIndex, adjustedPreviousIndex));
                changedIdxs.add(item);
                notifyParent = true;
            }
        });
        changes.forEachIdentityChange(function (record) {
            var item = record.item;
            changesArr.push(getUpdateChange(item, record.currentIndex));
            changedIdxs.add(item);
        });
        items.forEach(function (item, index) {
            if (!changedIdxs.has(item)) {
                changesArr.push(getUnchangedChange(item, index));
            }
        });
        return [changesArr, notifyParent];
        // ==========
        function getMoveChange(item, currentIndex, adjustedPreviousIndex) {
            return [
                2 /* move */,
                [item, currentIndex, adjustedPreviousIndex],
            ];
        }
        function getUpdateChange(item, currentIndex) {
            return [3 /* update */, [item, currentIndex]];
        }
        function getUnchangedChange(item, index) {
            return [4 /* context */, [item, index]];
        }
        function getInsertChange(item, currentIndex) {
            return [
                0 /* insert */,
                [item, currentIndex === null ? undefined : currentIndex],
            ];
        }
        function getRemoveChange(item, adjustedPreviousIndex) {
            return [
                1 /* remove */,
                [item, adjustedPreviousIndex === null ? undefined : adjustedPreviousIndex],
            ];
        }
    }

    function createListTemplateManager(config) {
        var templateSettings = config.templateSettings, renderSettings = config.renderSettings, trackBy = config.trackBy, iterableDiffers = config.iterableDiffers;
        var defaultStrategyName = renderSettings.defaultStrategyName, strategies = renderSettings.strategies, injectingViewCdRef = renderSettings.cdRef, patchZone = renderSettings.patchZone, parent = renderSettings.parent, eRef = renderSettings.eRef;
        var errorHandler = createErrorHandler(renderSettings.errorHandler);
        var strategyHandling$ = strategyHandling(defaultStrategyName, strategies);
        var differ = iterableDiffers.find([]).create(trackBy);
        //               type,  context
        var tNode = parent
            ? getTNode(injectingViewCdRef, eRef.nativeElement)
            : false;
        /* TODO (regarding createView): this is currently not in use. for the list-manager this would mean to provide
         functions for not only create. developers than should have to provide create, move, remove,... the whole thing.
         i don't know if this is the right decision for a first RC */
        var listViewHandler = getTemplateHandler(Object.assign(Object.assign({}, templateSettings), { initialTemplateRef: templateSettings.templateRef, patchZone: patchZone }));
        var viewContainerRef = templateSettings.viewContainerRef;
        var notifyParent = false;
        var changesArr;
        var partiallyFinished = false;
        return {
            nextStrategy: function (nextConfig) {
                strategyHandling$.next(nextConfig);
            },
            render: function (values$) {
                return values$.pipe(render());
            },
        };
        function render() {
            var count = 0;
            return function (o$) { return o$.pipe(
            // map iterable to latest diff
            operators.map(function (iterable) {
                if (partiallyFinished) {
                    var currentIterable = [];
                    for (var i = 0, ilen = viewContainerRef.length; i < ilen; i++) {
                        var viewRef = viewContainerRef.get(i);
                        currentIterable[i] = viewRef.context.$implicit;
                    }
                    differ.diff(currentIterable);
                }
                return {
                    changes: differ.diff(iterable),
                    items: iterable != null && Array.isArray(iterable) ? iterable : [],
                };
            }), operators.withLatestFrom(strategyHandling$.strategy$), 
            // Cancel old renders
            operators.switchMap(function (_a) {
                var _b = __read(_a, 2), _c = _b[0], changes = _c.changes, items = _c.items, strategy = _b[1];
                if (!changes) {
                    return rxjs.of([]);
                }
                var listChanges = listViewHandler.getListChanges(changes, items);
                changesArr = listChanges[0];
                var insertedOrRemoved = listChanges[1];
                var applyChanges$ = getObservablesFromChangesArray(changesArr, strategy, count);
                partiallyFinished = true;
                // @TODO we need to know if we need to notifyParent on move aswell
                notifyParent = insertedOrRemoved && parent;
                count = items.length;
                return rxjs.merge(rxjs.combineLatest(
                // emit after all changes are rendered
                applyChanges$.length > 0 ? applyChanges$ : [rxjs.of(items)]).pipe(operators.tap(function () { return (partiallyFinished = false); }), 
                // somehow this makes the strategySelect work
                notifyAllParentsIfNeeded(tNode, injectingViewCdRef, strategy, function () { return notifyParent; })), 
                // emit injectingParent if needed
                notifyInjectingParentIfNeeded(injectingViewCdRef, strategy, insertedOrRemoved).pipe(operators.ignoreElements())).pipe(operators.map(function () { return items; }), operators.catchError(function (e) {
                    partiallyFinished = false;
                    errorHandler.handleError(e);
                    return rxjs.of(e);
                }));
            })); };
        }
        /**
         * @internal
         *
         * returns an array of streams which process all of the view updates needed to reflect the latest diff to the
         * viewContainer.
         * I
         *
         * @param changes
         * @param strategy
         * @param count
         */
        function getObservablesFromChangesArray(changes, strategy, count) {
            return changes.length > 0
                ? changes.map(function (change) {
                    return onStrategy(change[1], strategy, function () {
                        var type = change[0];
                        var payload = change[1];
                        switch (type) {
                            case 0 /* insert */:
                                listViewHandler.insertView(payload[0], payload[1], count);
                                break;
                            case 2 /* move */:
                                listViewHandler.moveView(payload[2], payload[0], payload[1], count);
                                break;
                            case 1 /* remove */:
                                listViewHandler.removeView(payload[1]);
                                break;
                            case 3 /* update */:
                                listViewHandler.updateView(payload[0], payload[1], count);
                                break;
                            case 4 /* context */:
                                listViewHandler.updateUnchangedContext(payload[1], count);
                                break;
                        }
                    }, {}, function (e, v) { return toRenderError(e, v[0]); });
                })
                : [rxjs.of(null)];
        }
    }

    var computeFirst = function (_a) {
        var count = _a.count, index = _a.index;
        return index === 0;
    };
    var ɵ0$5 = computeFirst;
    var computeLast = function (_a) {
        var count = _a.count, index = _a.index;
        return index === count - 1;
    };
    var ɵ1$2 = computeLast;
    var computeEven = function (_a) {
        var count = _a.count, index = _a.index;
        return index % 2 === 0;
    };
    var ɵ2$2 = computeEven;
    var RxDefaultListViewContext = /** @class */ (function () {
        function RxDefaultListViewContext(item, customProps) {
            var _this = this;
            this.item = item;
            this._item = new rxjs.ReplaySubject(1);
            this.item$ = this._item.asObservable();
            this._context$ = new rxjs.BehaviorSubject({
                index: -1,
                count: -1,
            });
            this.select = function (props) {
                return _this.item$.pipe(operators.pluck.apply(void 0, __spread(props)));
            };
            // tslint:disable-next-line:no-unused-expression
            this.$implicit = item;
            if (customProps) {
                this.updateContext(customProps);
            }
        }
        Object.defineProperty(RxDefaultListViewContext.prototype, "$implicit", {
            get: function () {
                return this._$implicit;
            },
            set: function ($implicit) {
                this._$implicit = $implicit;
                this._item.next($implicit);
            },
            enumerable: false,
            configurable: true
        });
        Object.defineProperty(RxDefaultListViewContext.prototype, "$complete", {
            get: function () {
                return this._$complete;
            },
            enumerable: false,
            configurable: true
        });
        Object.defineProperty(RxDefaultListViewContext.prototype, "$error", {
            get: function () {
                return this._$error;
            },
            enumerable: false,
            configurable: true
        });
        Object.defineProperty(RxDefaultListViewContext.prototype, "$suspense", {
            get: function () {
                return this._$suspense;
            },
            enumerable: false,
            configurable: true
        });
        Object.defineProperty(RxDefaultListViewContext.prototype, "index", {
            get: function () {
                return this._context$.getValue().index;
            },
            enumerable: false,
            configurable: true
        });
        Object.defineProperty(RxDefaultListViewContext.prototype, "count", {
            get: function () {
                return this._context$.getValue().count;
            },
            enumerable: false,
            configurable: true
        });
        Object.defineProperty(RxDefaultListViewContext.prototype, "first", {
            get: function () {
                return computeFirst(this._context$.getValue());
            },
            enumerable: false,
            configurable: true
        });
        Object.defineProperty(RxDefaultListViewContext.prototype, "last", {
            get: function () {
                return computeLast(this._context$.getValue());
            },
            enumerable: false,
            configurable: true
        });
        Object.defineProperty(RxDefaultListViewContext.prototype, "even", {
            get: function () {
                return computeEven(this._context$.getValue());
            },
            enumerable: false,
            configurable: true
        });
        Object.defineProperty(RxDefaultListViewContext.prototype, "odd", {
            get: function () {
                return !this.even;
            },
            enumerable: false,
            configurable: true
        });
        Object.defineProperty(RxDefaultListViewContext.prototype, "index$", {
            get: function () {
                return this._context$.pipe(operators.pluck('index'), operators.distinctUntilChanged());
            },
            enumerable: false,
            configurable: true
        });
        Object.defineProperty(RxDefaultListViewContext.prototype, "count$", {
            get: function () {
                return this._context$.pipe(operators.pluck('count'), operators.distinctUntilChanged());
            },
            enumerable: false,
            configurable: true
        });
        Object.defineProperty(RxDefaultListViewContext.prototype, "first$", {
            get: function () {
                return this._context$.pipe(operators.map(computeFirst), operators.distinctUntilChanged());
            },
            enumerable: false,
            configurable: true
        });
        Object.defineProperty(RxDefaultListViewContext.prototype, "last$", {
            get: function () {
                return this._context$.pipe(operators.map(computeLast), operators.distinctUntilChanged());
            },
            enumerable: false,
            configurable: true
        });
        Object.defineProperty(RxDefaultListViewContext.prototype, "even$", {
            get: function () {
                return this._context$.pipe(operators.map(computeEven), operators.distinctUntilChanged());
            },
            enumerable: false,
            configurable: true
        });
        Object.defineProperty(RxDefaultListViewContext.prototype, "odd$", {
            get: function () {
                return this.even$.pipe(operators.map(function (even) { return !even; }));
            },
            enumerable: false,
            configurable: true
        });
        RxDefaultListViewContext.prototype.updateContext = function (newProps) {
            this._context$.next(Object.assign(Object.assign({}, this._context$.getValue()), newProps));
        };
        return RxDefaultListViewContext;
    }());

    function isNumeric(val) {
        // parseFloat NaNs numeric-cast false positives (null|true|false|"")
        // ...but misinterprets leading-number strings, particularly hex literals ("0x...")
        // subtraction forces infinities to NaN
        // adding 1 corrects loss of precision from parseFloat (#15100)
        return !Array.isArray(val) && val - parseFloat(val) + 1 >= 0;
    }
    function isScheduler(value) {
        return value && typeof value.schedule === 'function';
    }

    /**
     * Creates an Observable that emits sequential numbers every specified
     * interval of time, on a specified {@link SchedulerLike}.
     *
     * <span class="informal">Emits incremental numbers periodically in time.
     * </span>
     *
     * ![](interval.png)
     *
     * `interval` returns an Observable that emits an infinite sequence of
     * ascending integers, with a constant interval of time of your choosing
     * between those emissions. The first emission is not sent immediately, but
     * only after the first period has passed. By default, this operator uses the
     * `async` {@link SchedulerLike} to provide a notion of time, but you may pass any
     * {@link SchedulerLike} to it.
     *
     * ## Example
     * Emits ascending numbers, one every second (1000ms) up to the number 3
     * ```ts
     * import { interval } from 'rxjs';
     * import { take } from 'rxjs/operators';
     *
     * const numbers = interval(1000);
     *
     * const takeFourNumbers = numbers.pipe(take(4));
     *
     * takeFourNumbers.subscribe(x => console.log('Next: ', x));
     *
     * // Logs:
     * // Next: 0
     * // Next: 1
     * // Next: 2
     * // Next: 3
     * ```
     *
     * @see {@link timer}
     * @see {@link delay}
     *
     * @param {number} [period=0] The interval size in milliseconds (by default)
     * or the time unit determined by the scheduler's clock.
     * @param {SchedulerLike} [scheduler=async] The {@link SchedulerLike} to use for scheduling
     * the emission of values, and providing a notion of "time".
     * @return {Observable} An Observable that emits a sequential number each time
     * interval.
     * @static true
     * @name interval
     * @owner Observable
     */
    function interval(period, scheduler) {
        if (period === void 0) { period = 0; }
        if (scheduler === void 0) { scheduler = async; }
        if (!isNumeric(period) || period < 0) {
            period = 0;
        }
        if (!scheduler || typeof scheduler.schedule !== 'function') {
            scheduler = async;
        }
        return new rxjs.Observable(function (subscriber) {
            subscriber.add(scheduler.schedule(dispatch, period, { subscriber: subscriber, counter: 0, period: period }));
            return subscriber;
        });
    }
    function dispatch(state) {
        var subscriber = state.subscriber, counter = state.counter, period = state.period;
        subscriber.next(counter);
        this.schedule({ subscriber: subscriber, counter: counter + 1, period: period }, period);
    }

    /**
     * Creates an Observable that starts emitting after an `dueTime` and
     * emits ever increasing numbers after each `period` of time thereafter.
     *
     * <span class="informal">Its like {@link index/interval}, but you can specify when
     * should the emissions start.</span>
     *
     * ![](timer.png)
     *
     * `timer` returns an Observable that emits an infinite sequence of ascending
     * integers, with a constant interval of time, `period` of your choosing
     * between those emissions. The first emission happens after the specified
     * `dueTime`. The initial delay may be a `Date`. By default, this
     * operator uses the {@link asyncScheduler} {@link SchedulerLike} to provide a notion of time, but you
     * may pass any {@link SchedulerLike} to it. If `period` is not specified, the output
     * Observable emits only one value, `0`. Otherwise, it emits an infinite
     * sequence.
     *
     * ## Examples
     * ### Emits ascending numbers, one every second (1000ms), starting after 3 seconds
     * ```ts
     * import { timer } from 'rxjs';
     *
     * const numbers = timer(3000, 1000);
     * numbers.subscribe(x => console.log(x));
     * ```
     *
     * ### Emits one number after five seconds
     * ```ts
     * import { timer } from 'rxjs';
     *
     * const numbers = timer(5000);
     * numbers.subscribe(x => console.log(x));
     * ```
     * @see {@link index/interval}
     * @see {@link delay}
     *
     * @param {number|Date} [dueTime] The initial delay time specified as a Date object or as an integer denoting
     * milliseconds to wait before emitting the first value of 0`.
     * @param {number|SchedulerLike} [periodOrScheduler] The period of time between emissions of the
     * subsequent numbers.
     * @param {SchedulerLike} [scheduler=async] The {@link SchedulerLike} to use for scheduling
     * the emission of values, and providing a notion of "time".
     * @return {Observable} An Observable that emits a `0` after the
     * `dueTime` and ever increasing numbers after each `period` of time
     * thereafter.
     * @static true
     * @name timer
     * @owner Observable
     */
    function timer(dueTime, periodOrScheduler, scheduler) {
        if (dueTime === void 0) { dueTime = 0; }
        var period = -1;
        if (isNumeric(periodOrScheduler)) {
            period = (Number(periodOrScheduler) < 1 && 1) || Number(periodOrScheduler);
        }
        else if (isScheduler(periodOrScheduler)) {
            scheduler = periodOrScheduler;
        }
        if (!isScheduler(scheduler)) {
            scheduler = async;
        }
        return new rxjs.Observable(function (subscriber) {
            var due = isNumeric(dueTime)
                ? dueTime
                : +dueTime - scheduler.now();
            return scheduler.schedule(dispatch$1, due, {
                index: 0,
                period: period,
                subscriber: subscriber,
            });
        });
    }
    function dispatch$1(state) {
        var index = state.index, period = state.period, subscriber = state.subscriber;
        subscriber.next(index);
        if (subscriber.closed) {
            return;
        }
        else if (period === -1) {
            return subscriber.complete();
        }
        state.index = index + 1;
        this.schedule(state, period);
    }

    // @ts-ignore
    var isFunction = function (fn) { return typeof fn === 'function'; };
    var ɵ0$6 = isFunction;
    var isArray = Array.isArray;
    var ɵ1$3 = function () { return Object.prototype.toString; };
    var toString = (ɵ1$3)();
    /* tslint:enable:max-line-length */
    function fromEvent(target, eventName, options, resultSelector) {
        if (isFunction(options)) {
            // DEPRECATED PATH
            // @ts-ignore
            resultSelector = options;
            options = undefined;
        }
        if (resultSelector) {
            // DEPRECATED PATH
            return fromEvent(target, eventName, options).pipe(operators.map(function (args) { return isArray(args) ? resultSelector.apply(void 0, __spread(args)) : resultSelector(args); }));
        }
        return new rxjs.Observable(function (subscriber) {
            function handler(e) {
                if (arguments.length > 1) {
                    subscriber.next(Array.prototype.slice.call(arguments));
                }
                else {
                    subscriber.next(e);
                }
            }
            setupSubscription(target, eventName, handler, subscriber, options);
        });
    }
    function setupSubscription(sourceObj, eventName, handler, subscriber, options) {
        var unsubscribe;
        if (isEventTarget(sourceObj)) {
            var source_1 = sourceObj;
            getZoneUnPatchedApi('addEventListener', sourceObj).call(sourceObj, eventName, handler, options);
            unsubscribe = function () { return getZoneUnPatchedApi('removeEventListener', source_1).call(source_1, eventName, handler, options); };
        }
        else if (isJQueryStyleEventEmitter(sourceObj)) {
            var source_2 = sourceObj;
            sourceObj.on(eventName, handler);
            unsubscribe = function () { return source_2.off(eventName, handler); };
        }
        else if (isNodeStyleEventEmitter(sourceObj)) {
            var source_3 = sourceObj;
            sourceObj.addListener(eventName, handler);
            unsubscribe = function () { return source_3.removeListener(eventName, handler); };
        }
        else if (sourceObj && sourceObj.length) {
            for (var i = 0, len = sourceObj.length; i < len; i++) {
                setupSubscription(sourceObj[i], eventName, handler, subscriber, options);
            }
        }
        else {
            throw new TypeError('Invalid event target');
        }
        subscriber.add(unsubscribe);
    }
    function isNodeStyleEventEmitter(sourceObj) {
        return (sourceObj &&
            typeof sourceObj.addListener === 'function' &&
            typeof sourceObj.removeListener === 'function');
    }
    function isJQueryStyleEventEmitter(sourceObj) {
        return (sourceObj &&
            typeof sourceObj.on === 'function' &&
            typeof sourceObj.off === 'function');
    }
    function isEventTarget(sourceObj) {
        return (sourceObj &&
            typeof sourceObj.addEventListener === 'function' &&
            typeof sourceObj.removeEventListener === 'function');
    }

    // Standard Events
    // Focus Events
    var focusEvents = ['blur', 'focus', 'focusin', 'focusout'];
    /**
     * Mouse Events
     * (MouseEvent)[https://developer.mozilla.org/en-US/docs/Web/API/MouseEvent]
     */
    var mouseEvents = [
        'mousedown',
        'dblclick',
        'mouseenter',
        'mouseleave',
        'mousemove',
        'mouseout',
        'mouseover',
        'mouseup',
        'click',
    ];
    /**
     * Wheel Events
     * (WheelEvent)[https://developer.mozilla.org/en-US/docs/Web/API/WheelEvent]
     */
    var wheelEvents = [
        // (WheelEvent)[https://developer.mozilla.org/en-US/docs/Web/API/WheelEvent]
        'wheel',
        'mousewheel',
    ];
    // Input Events
    var inputEvents = [
        'input',
        'invalid',
        'change',
        'reset',
        'select',
        'submit',
    ];
    /**
     * @deprecated
     */
    var formControlsEvents = inputEvents;
    // [KeyboardEvent](https://developer.mozilla.org/en-US/docs/Web/API/KeyboardEvent)
    var keyboardEvents = [
        'keydown',
        'keypress',
        'keyup',
    ];
    // [VR]()
    var vrEvents = [
        'vrdisplayactivate',
        'vrdisplayblur',
        'vrdisplayconnect',
        'vrdisplaydeactivate',
        'vrdisplaydisconnect',
        'vrdisplayfocus',
        'vrdisplaypointerrestricted',
        'vrdisplaypointerunrestricted',
        'vrdisplaypresentchange',
    ];
    // [MSGesture]()
    var mSGestureEvents = [
        'MSGestureChange',
        'MSGestureDoubleTap',
        'MSGestureEnd',
        'MSGestureHold',
        'MSGestureStart',
        'MSGestureTap',
        'MSInertiaStart',
        'MSPointerCancel',
        'MSPointerDown',
        'MSPointerEnter',
        'MSPointerLeave',
        'MSPointerMove',
        'MSPointerOut',
        'MSPointerOver',
        'MSPointerUp',
    ];
    // [xPrint](https://developer.mozilla.org/en-US/docs/Web/API/Window/afterprint_event)
    var printEvents = [
        'afterprint',
        'beforeprint',
    ];
    // [network]()
    var networkEvents = ['offline', 'online'];
    // [network]()
    var audioEvents = [
        'canplay',
        'canplaythrough',
        'pause',
        'play',
        'playing',
        'volumechange',
    ];
    // Composition Events
    var compositionEvents = [
        'compositionstart',
        'compositionupdate',
        'compositionend',
    ];
    /**
     * Touch Events
     * [PointerEvent](https://developer.mozilla.org/en-US/docs/Web/API/PointerEvent)
     * [TouchEvent](https://developer.mozilla.org/en-US/docs/Web/API/TouchEvent)
     */
    var touchEvents = [
        // [PointerEvent](https://developer.mozilla.org/en-US/docs/Web/API/PointerEvent)
        'pointerover',
        'pointerenter',
        'pointerdown',
        'pointermove',
        // 'pointerrawupdate',
        'pointerup',
        'pointercancel',
        'pointerout',
        'pointerleave',
        'gotpointercapture',
        'lostpointercapture',
        'pointerup',
        // [TouchEvent](https://developer.mozilla.org/en-US/docs/Web/API/TouchEvent)
        'touchstart',
        'touchend',
        'touchmove',
        'touchcancel',
        // [DragEvent](https://developer.mozilla.org/en-US/docs/Web/API/DragEvent)
        'drag',
        'dragend',
        'dragexit',
        'dragenter',
        'dragleave',
        'dragover',
        'dragstart',
        'drop',
    ];
    var globalEvents = [
        'contextmenu',
        'resize',
        'scroll',
        'abort',
        'load',
        'loadeddata',
        'loadedmetadata',
        'loadstart',
        'unload',
        'error',
    ];
    var websocketEvents = [
        'close',
        'error',
        'message',
        'open',
    ];
    /**
     * Basic XHR Events
     * [Load](https://developer.mozilla.org/en-US/docs/Web/API/XMLHttpRequest/load_event)
     * [Error](https://developer.mozilla.org/en-US/docs/Web/API/XMLHttpRequest/error_event)
     * There are more events you may want to unpatch https://developer.mozilla.org/en-US/docs/Web/API/XMLHttpRequest
     */
    // XHREvent (group is here to make it easier to target XHR in angular applications. It contains redundant events e.g. 'error')
    var xhrEvents = ['load', 'error'];
    // @TODO
    var windowEvents = [
        'compassneedscalibration',
        'durationchange',
        'emptied',
        'ended',
        'orientationchange',
        'ratechange',
        'seeked',
        'seeking',
        'stalled',
        'suspend',
        'timeupdate',
        'waiting',
    ];
    /**
     * All Events combined
     */
    var allEvents = Array.from(new Set(__spread(focusEvents, mouseEvents, wheelEvents, inputEvents, keyboardEvents, inputEvents, vrEvents, mSGestureEvents, printEvents, networkEvents, audioEvents, compositionEvents, touchEvents, globalEvents, websocketEvents, xhrEvents, windowEvents)));

    var zoneGlobalDisableConfigurationsKeys = [
        'EventEmitter',
        'fs',
        'node_timers',
        'nextTick',
        'crypto',
        'defineProperty',
        'registerElement',
        'EventTargetLegacy',
        'timers',
        'requestAnimationFrame',
        'blocking',
        'EventTarget',
        'FileReader',
        'MutationObserver',
        'IntersectionObserver',
        'on_property',
        'customElements',
        'XHR',
        'geolocation',
        'canvas',
        'ZoneAwarePromise',
    ];
    var zoneGlobalEventsConfigurationsKeys = [
        'UNPATCHED_EVENTS',
        'PASSIVE_EVENTS',
    ];
    var zoneGlobalSettingsConfigurationsKeys = [
        'DISABLE_WRAPPING_UNCAUGHT_PROMISE_REJECTION',
    ];
    var zoneRuntimeConfigurationsKeys = [
        'ignoreConsoleErrorUncaughtError',
    ];
    var zoneTestDisableConfigurationsKeys = [
        'jasmine',
        'mocha',
        'jest',
    ];
    var zoneTestSettingsConfigurationsKeys = [
        'fakeAsyncDisablePatchingClock',
        'fakeAsyncAutoFakeAsyncWhenClockPatched',
        'supportWaitUnResolvedChainedPromise',
    ];

    var convenienceMethods = function (config) { return ({
        unpatchXHR: function () {
            config.global.disable.XHR();
            config.events.disable.UNPATCHED_EVENTS(__spread(xhrEvents));
        },
        useUnpatchedPassiveScrollEvents: function () {
            config.events.disable.PASSIVE_EVENTS(['scroll']);
            config.events.disable.UNPATCHED_EVENTS(['scroll']);
        },
    }); };

    var zoneDisable = '__Zone_disable_';
    var zoneSymbol = '__zone_symbol__';
    /**
     * https://angular.io/guide/zone#setting-up-zonejs
     **/
    function assertZoneConfig() {
        if (window.Zone !== undefined) {
            // @TODO link to docs
            console.error('zone-flags file needs to get imported before zone.js');
        }
    }
    var addDisableFlag = function (prop) {
        var _a;
        return (_a = {},
            _a[prop] = function () {
                assertZoneConfig();
                return (window[zoneDisable + prop] = true);
            },
            _a);
    };
    var ɵ0$7 = addDisableFlag;
    var addSymbolFlag = function (prop) {
        var _a;
        return (_a = {},
            _a[prop] = function () {
                assertZoneConfig();
                return (window[zoneSymbol + prop] = true);
            },
            _a);
    };
    var ɵ1$4 = addSymbolFlag;
    var addArraySymbolFlag = function (prop) {
        var _a;
        return (_a = {},
            _a[prop] = function (eventNames) {
                assertZoneConfig();
                var w = window;
                return (w[zoneSymbol + prop] = __spread((Array.isArray(w[zoneSymbol + prop]) ? w[zoneSymbol + prop] : []), eventNames));
            },
            _a);
    };
    var ɵ2$3 = addArraySymbolFlag;
    var reduceToObject = function (methodsArray) {
        return methodsArray.reduce(function (map, item) { return (Object.assign(Object.assign({}, map), item)); }, {});
    };
    var ɵ3$2 = reduceToObject;
    /**
     * factory function to create a `ZoneConfig` object.
     *
     * @Example
     * import { globalEvents,xhrEvent, zoneConfig} from '@rx-angular/cdk/zone-flags';
     *
     * const zoneConfig = createZoneFlagsConfigurator();
     *
     * zoneConfig.global.disable.requestAnimationFrame();
     * zoneConfig.global.disable.timers();
     * zoneConfig.events.disable.UNPATCHED_EVENTS([...globalEvents, ...xhrEvent]);
     *
     */
    function createZoneFlagsConfigurator() {
        var cfg = i0.ɵglobal;
        var configProps = __spread(__spread(zoneGlobalDisableConfigurationsKeys, zoneTestDisableConfigurationsKeys).map(function (prop) { return zoneDisable + prop; }), __spread(zoneGlobalSettingsConfigurationsKeys, zoneTestSettingsConfigurationsKeys, zoneGlobalEventsConfigurationsKeys, zoneRuntimeConfigurationsKeys).map(function (prop) { return zoneSymbol + prop; }));
        // append as global method for easy debugging
        cfg.__rxa_zone_config__log = function () {
            configProps.forEach(function (flag) {
                // tslint:disable-next-line:no-unused-expression
                cfg[flag] && console.log(flag, cfg[flag]);
            });
        };
        var zoneConfigObj = {
            global: {
                disable: reduceToObject(__spread(zoneGlobalDisableConfigurationsKeys.map(addDisableFlag), zoneGlobalSettingsConfigurationsKeys.map(addSymbolFlag))),
            },
            test: {
                disable: reduceToObject(__spread(zoneTestDisableConfigurationsKeys.map(addDisableFlag), zoneTestSettingsConfigurationsKeys.map(addSymbolFlag))),
            },
            events: {
                disable: reduceToObject(zoneGlobalEventsConfigurationsKeys.map(addArraySymbolFlag)),
            },
            runtime: {
                disable: reduceToObject(zoneRuntimeConfigurationsKeys.map(addSymbolFlag)),
            },
        };
        return Object.assign(Object.assign({}, zoneConfigObj), convenienceMethods(zoneConfigObj));
    }
    /**
     * An object for typed zone-flags configuration.
     *
     * @Example
     *
     * create file `zone-flags.ts` parallel to your `polyfills.ts` and insert following content:
     * ```typescript
     * import { globalEvents, xhrEvent, zoneConfig} from '@rx-angular/cdk/zone-flags';
     *
     * zoneConfig.global.disable.requestAnimationFrame();
     * zoneConfig.global.disable.timers();
     * zoneConfig.events.disable.UNPATCHED_EVENTS([...globalEvents, ...xhrEvent]);
     * ```
     * In `polyfills.ts` above the zone import, import `zone-flags.ts`
     *
     * ```typescript
     * import './zone-flags';
     * // Zone JS is required by default for Angular itself.
     * import 'zone.js/dist/zone';
     * ```
     *
     */
    var zoneConfig = createZoneFlagsConfigurator();

    /**
     * Generated bundle index. Do not edit.
     */

    exports.Promise = Promise$1;
    exports.RX_ANGULAR_CONFIG = RX_ANGULAR_CONFIG;
    exports.RX_ANGULAR_DEFAULTS = RX_ANGULAR_DEFAULTS;
    exports.RX_CONCURRENT_STRATEGIES = RX_CONCURRENT_STRATEGIES;
    exports.RX_NATIVE_STRATEGIES = RX_NATIVE_STRATEGIES;
    exports.RxDefaultListViewContext = RxDefaultListViewContext;
    exports.RxStrategyProvider = RxStrategyProvider;
    exports.allEvents = allEvents;
    exports.animationFrame = animationFrame;
    exports.asap = asap;
    exports.asyncScheduler = async;
    exports.audioEvents = audioEvents;
    exports.cancelAnimationFrame = cancelAnimationFrame$1;
    exports.clearInterval = clearInterval;
    exports.clearTimeout = clearTimeout$1;
    exports.coalesceWith = coalesceWith;
    exports.coerceDistinctObservable = coerceDistinctObservable;
    exports.coerceDistinctWith = coerceDistinctWith;
    exports.coerceObservable = coerceObservable;
    exports.coerceObservableWith = coerceObservableWith;
    exports.compositionEvents = compositionEvents;
    exports.createListTemplateManager = createListTemplateManager;
    exports.createTemplateManager = createTemplateManager;
    exports.focusEvents = focusEvents;
    exports.formControlsEvents = formControlsEvents;
    exports.fromEvent = fromEvent;
    exports.getZoneUnPatchedApi = getZoneUnPatchedApi;
    exports.globalEvents = globalEvents;
    exports.hotFlatten = hotFlatten;
    exports.inputEvents = inputEvents;
    exports.interval = interval;
    exports.keyboardEvents = keyboardEvents;
    exports.mSGestureEvents = mSGestureEvents;
    exports.mergeDefaultConfig = mergeDefaultConfig;
    exports.mouseEvents = mouseEvents;
    exports.networkEvents = networkEvents;
    exports.onStrategy = onStrategy;
    exports.printEvents = printEvents;
    exports.queue = queue;
    exports.requestAnimationFrame = requestAnimationFrame$1;
    exports.rxMaterialize = rxMaterialize;
    exports.setInterval = setInterval;
    exports.setTimeout = setTimeout$1;
    exports.strategyHandling = strategyHandling;
    exports.templateHandling = templateHandling;
    exports.templateNotifier = templateNotifier;
    exports.templateTriggerHandling = templateTriggerHandling;
    exports.timer = timer;
    exports.toRxCompleteNotification = toRxCompleteNotification;
    exports.toRxErrorNotification = toRxErrorNotification;
    exports.toRxSuspenseNotification = toRxSuspenseNotification;
    exports.touchEvents = touchEvents;
    exports.unpatchAddEventListener = unpatchAddEventListener;
    exports.vrEvents = vrEvents;
    exports.websocketEvents = websocketEvents;
    exports.wheelEvents = wheelEvents;
    exports.windowEvents = windowEvents;
    exports.xhrEvents = xhrEvents;
    exports.zoneConfig = zoneConfig;

    Object.defineProperty(exports, '__esModule', { value: true });

})));
//# sourceMappingURL=cdk.umd.js.map
