import { ChangeDetectorRef, Type } from '@angular/core';
import { Notification, Observable } from 'rxjs';
export declare type coalescingObj = Record<string | number | symbol, unknown> | Type<unknown> | object;
export interface RxCoalescingOptions {
    scope?: coalescingObj;
}
export declare enum RxNotificationKind {
    suspense = "suspense",
    next = "next",
    error = "error",
    complete = "complete"
}
export declare type RxNotificationValue = 'value' | 'hasValue';
export declare type RxNextNotification<T> = Pick<Notification<T>, RxNotificationValue> & {
    kind: RxNotificationKind;
} & {
    error: boolean;
} & {
    complete: boolean;
};
export declare type RxSuspenseNotification<T> = Pick<Notification<T>, RxNotificationValue> & {
    kind: RxNotificationKind.suspense;
} & {
    error: false;
} & {
    complete: false;
};
export declare type RxErrorNotification<T> = Pick<Notification<T>, RxNotificationValue> & {
    kind: RxNotificationKind.error;
} & {
    error: any;
} & {
    complete: false;
};
export declare type RxCompleteNotification<T> = Pick<Notification<T>, RxNotificationValue> & {
    kind: RxNotificationKind.complete;
} & {
    complete: boolean;
} & {
    error: false;
};
export declare type RxNotification<T> = RxNextNotification<T> | RxSuspenseNotification<T> | RxErrorNotification<T> | RxCompleteNotification<T>;
export declare type RxRenderWork = <T = unknown>(cdRef: ChangeDetectorRef, scope?: coalescingObj, notification?: RxNotification<T>) => void;
export declare type RxRenderBehavior = <T = unknown>(work: any, scope?: coalescingObj) => (o: Observable<T>) => Observable<T>;
export interface RxStrategyCredentials<S = string> {
    name: S;
    work: RxRenderWork;
    behavior: RxRenderBehavior;
}
export declare type RxCustomStrategyCredentials<T extends string> = Record<T, RxStrategyCredentials>;
export declare type RxNativeStrategyNames = 'native' | 'local' | 'global' | 'noop';
export declare type RxConcurrentStrategyNames = 'noPriority' | 'immediate' | 'userBlocking' | 'normal' | 'low' | 'idle';
export declare type RxDefaultStrategyNames = RxNativeStrategyNames | RxConcurrentStrategyNames;
export declare type RxStrategyNames<T> = RxDefaultStrategyNames | T;
export declare type RxStrategies<T extends string> = RxCustomStrategyCredentials<RxStrategyNames<T>>;
export interface RxAngularConfig<T extends string> {
    primaryStrategy?: RxStrategyNames<T>;
    customStrategies?: RxCustomStrategyCredentials<T>;
    patchZone?: boolean;
}
